import sys
import unittest
import importlib

import asn1tools

sys.path.append('tests/files')
sys.path.append('tests/files/etsi')
sys.path.append('tests/files/ietf')
sys.path.append('tests/files/3gpp')


class Asn1ToolsParseTest(unittest.TestCase):

    maxDiff = None

    def parse_and_verify(self, module, path='.'):
        asn_path = 'tests/files/' + path + '/' + module + '.asn'
        actual = asn1tools.parse_files(asn_path)

        # from pprint import pformat
        #
        # py_path = 'tests/files/' + path + '/' + module + '.py'
        #
        # with open(py_path, 'w') as fout:
        #     fout.write('EXPECTED = ' + pformat(actual))

        module = importlib.import_module(module)
        self.assertEqual(actual, module.EXPECTED)

    def test_parse_foo(self):
        self.parse_and_verify('foo')

    def test_parse_bar(self):
        self.parse_and_verify('bar')

    def test_parse_all_types(self):
        self.parse_and_verify('all_types')

    def test_parse_extensibility_implied(self):
        self.parse_and_verify('extensibility_implied')

    def test_parse_all_types_automatic_tags(self):
        self.parse_and_verify('all_types_automatic_tags')

    def test_parse_module_tags_explicit(self):
        self.parse_and_verify('module_tags_explicit')

    def test_parse_module_tags_implicit(self):
        self.parse_and_verify('module_tags_implicit')

    def test_parse_module_tags_automatic(self):
        self.parse_and_verify('module_tags_automatic')

    def test_parse_information_object(self):
        self.parse_and_verify('information_object')

    def test_parse_x680(self):
        self.parse_and_verify('x680')

    def test_parse_x691_a1(self):
        self.parse_and_verify('x691_a1')

    def test_parse_x691_a2(self):
        with self.assertRaises(AssertionError):
            self.parse_and_verify('x691_a2')

    def test_parse_x691_a3(self):
        self.parse_and_verify('x691_a3')

    def test_parse_x691_a4(self):
        self.parse_and_verify('x691_a4')

    def test_parse_zforce(self):
        self.parse_and_verify('zforce')

    def test_parse_rrc_8_6_0(self):
        self.parse_and_verify('rrc_8_6_0', '3gpp')

    def test_parse_rrc_14_4_0(self):
        self.parse_and_verify('rrc_14_4_0', '3gpp')

    def test_parse_s1ap_14_4_0(self):
        self.parse_and_verify('s1ap_14_4_0', '3gpp')

    def test_parse_lpp_14_3_0(self):
        self.parse_and_verify('lpp_14_3_0', '3gpp')

    def test_parse_rfc1155(self):
        self.parse_and_verify('rfc1155', 'ietf')

    def test_parse_rfc1157(self):
        self.parse_and_verify('rfc1157', 'ietf')

    def test_parse_rfc2986(self):
        self.parse_and_verify('rfc2986', 'ietf')

    def test_parse_rfc3161(self):
        self.parse_and_verify('rfc3161', 'ietf')

    def test_parse_rfc3279(self):
        self.parse_and_verify('rfc3279', 'ietf')

    def test_parse_rfc3281(self):
        self.parse_and_verify('rfc3281', 'ietf')

    def test_parse_rfc3447(self):
        self.parse_and_verify('rfc3447', 'ietf')

    def test_parse_rfc3852(self):
        self.parse_and_verify('rfc3852', 'ietf')

    def test_parse_rfc4210(self):
        self.parse_and_verify('rfc4210', 'ietf')

    def test_parse_rfc4211(self):
        self.parse_and_verify('rfc4211', 'ietf')

    def test_parse_rfc4511(self):
        self.parse_and_verify('rfc4511', 'ietf')

    def test_parse_rfc5084(self):
        self.parse_and_verify('rfc5084', 'ietf')

    def test_parse_rfc5280(self):
        self.parse_and_verify('rfc5280', 'ietf')

    def test_parse_rfc5280_modified(self):
        self.parse_and_verify('rfc5280_modified', 'ietf')

    def test_parse_etsi_cam_pdu_descriptions_1_3_2(self):
        self.parse_and_verify('cam_pdu_descriptions_1_3_2', 'etsi')

    def test_parse_etsi_its_container_1_2_1(self):
        self.parse_and_verify('its_container_1_2_1', 'etsi')

    def test_parse_enumerated(self):
        self.parse_and_verify('enumerated')

    def test_parse_imports_global_module_reference(self):
        actual = asn1tools.parse_string('A DEFINITIONS ::= BEGIN '
                                        'IMPORTS '
                                        'a FROM B '
                                        'c, d FROM E global-module-reference '
                                        'f, g FROM H {iso(1)}; '
                                        'END')

        expected = {
            'A': {
                'extensibility-implied': False,
                'imports': {
                    'B': ['a'],
                    'E': ['c', 'd'],
                    'H': ['f', 'g']
                },
                'object-classes': {},
                'object-sets': {},
                'types': {},
                'values': {}
            }
        }

        self.assertEqual(actual, expected)

    def test_parse_imports_single_value_reference(self):
        """Test that a value reference, in this test 'c', is not parsed as an
        assignmed identifer, but an imported value from 'D'.

        """

        actual = asn1tools.parse_string('A DEFINITIONS ::= BEGIN '
                                        'IMPORTS '
                                        'A FROM B '
                                        'c FROM D; '
                                        'END')

        expected = {
            'A': {
                'extensibility-implied': False,
                'imports': {
                    'B': ['A'],
                    'D': ['c']
                },
                'object-classes': {},
                'object-sets': {},
                'types': {},
                'values': {}
            }
        }

        self.assertEqual(actual, expected)

    def test_parse_keyword_in_type_name(self):
        actual = asn1tools.parse_string('A DEFINITIONS ::= BEGIN '
                                        'ENDa ::= INTEGER '
                                        'END')

        expected = {
            'A': {
                'extensibility-implied': False,
                'imports': {},
                'object-classes': {},
                'object-sets': {},
                'types': {'ENDa': {'type': 'INTEGER'}},
                'values': {}
            }
        }

        self.assertEqual(actual, expected)

    def test_parse_error_empty_string(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('')

        self.assertEqual(str(cm.exception),
                         "Invalid ASN.1 syntax at line 1, column 1: '>!<': "
                         "Expected modulereference.")

    def test_parse_error_begin_missing(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= END')

        self.assertEqual(str(cm.exception),
                         "Invalid ASN.1 syntax at line 1, column 19: "
                         "'A DEFINITIONS ::= >!<END': Expected BEGIN.")

    def test_parse_error_end_missing(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN')

        self.assertEqual(str(cm.exception),
                         "Invalid ASN.1 syntax at line 1, column 24: "
                         "'A DEFINITIONS ::= BEGIN>!<': Expected END.")

    def test_parse_error_type_assignment_missing_assignment(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN A END')

        self.assertEqual(str(cm.exception),
                         "Invalid ASN.1 syntax at line 1, column 27: "
                         "'A DEFINITIONS ::= BEGIN A >!<END': "
                         "Expected ::=.")

    def test_parse_error_value_assignment_missing_assignment(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN a INTEGER END')

        self.assertEqual(str(cm.exception),
                         "Invalid ASN.1 syntax at line 1, column 35: "
                         "'A DEFINITIONS ::= BEGIN a INTEGER >!<END': "
                         "Expected ::=.")

    def test_parse_error_sequence_missing_type(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN'
                                   '  A ::= SEQUENCE { a } '
                                   'END')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 1, column 45: 'A DEFINITIONS ::= BEGIN "
            " A ::= SEQUENCE { a >!<} END': Expected Type.")

    def test_parse_error_sequence_missing_member_name(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN'
                                   '  A ::= SEQUENCE { A } '
                                   'END')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 1, column 43: 'A DEFINITIONS ::= "
            "BEGIN  A ::= SEQUENCE { >!<A } END': Expected \"}\".")

    def test_parse_error_definitive_identifier(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A {} DEFINITIONS ::= BEGIN '
                                   'END')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 1, column 4: 'A {>!<} DEFINITIONS "
            "::= BEGIN END': Expected {{identifier Suppress:(\"(\") - "
            "definitiveNumberForm - Suppress:(\")\")} | identifier | "
            "definitiveNumberForm}.")

    def test_parse_error_missing_union_member_beginning(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN '
                                   'B ::= INTEGER (| SIZE (1))'
                                   'END')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 1, column 40: 'A DEFINITIONS ::= BEGIN "
            "B ::= INTEGER (>!<| SIZE (1))END': Expected one or more constraints.")

    def test_parse_error_missing_union_member_middle(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN '
                                   'B ::= INTEGER (SIZE (1) | | (0))'
                                   'END')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 1, column 49: \'A DEFINITIONS "
            "::= BEGIN B ::= INTEGER (SIZE (1) >!<| | (0))END\': Expected \")\".")

    def test_parse_error_missing_union_member_end(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN '
                                   'B ::= INTEGER (SIZE (1) |)'
                                   'END')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 1, column 49: \'A DEFINITIONS "
            "::= BEGIN B ::= INTEGER (SIZE (1) >!<|)END\': Expected \")\".")

    def test_parse_error_size_constraint_missing_parentheses(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN '
                                   'B ::= INTEGER (SIZE 1)'
                                   'END')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 1, column 45: \'A DEFINITIONS ::= "
            "BEGIN B ::= INTEGER (SIZE >!<1)END\': Expected \"(\".")

    def test_parse_error_size_constraint_missing_size(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN '
                                   'B ::= INTEGER (SIZE ())'
                                   'END')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 1, column 46: 'A DEFINITIONS ::= "
            "BEGIN B ::= INTEGER (SIZE (>!<))END': Expected one or more "
            "constraints.")

    def test_parse_error_tag_class_number_missing(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN '
                                   'B ::= [] INTEGER '
                                   'END')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 1, column 32: 'A DEFINITIONS "
            "::= BEGIN B ::= [>!<] INTEGER END': Expected ClassNumber.")

    def test_parse_error_missing_type(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN '
                                   'B ::= '
                                   'END')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 1, column 31: 'A DEFINITIONS ::= BEGIN "
            "B ::= >!<END': Expected Type.")

    def test_parse_error_end_missing_with_comments(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS -- g -- \n'
                                   '-- hhhh\n'
                                   '::= BEGIN ')

        self.assertEqual(str(cm.exception),
                         "Invalid ASN.1 syntax at line 3, column 11: "
                         "'::= BEGIN >!<': Expected END.")

    def test_parse_error_late_extension_additions(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN '
                                   'Foo ::= SEQUENCE { '
                                   'a BOOLEAN, '
                                   '..., '
                                   '..., '
                                   '[[ '
                                   'c BOOLEAN '
                                   ']] '
                                   '} '
                                   'END')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 1, column 63: \'A DEFINITIONS ::= "
            "BEGIN Foo ::= SEQUENCE { a BOOLEAN, ..., ...>!<, [[ c BOOLEAN ]] "
            "} END\': Expected \"}\".")

    def test_parse_error_too_many_extension_markers(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN '
                                   'Foo ::= SEQUENCE { '
                                   'a BOOLEAN, '
                                   '..., '
                                   '[[ '
                                   'b BOOLEAN '
                                   ']], '
                                   '[[ '
                                   'c BOOLEAN '
                                   ']], '
                                   '..., '
                                   'd BOOLEAN, '
                                   '... '
                                   '} '
                                   'END')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 1, column 108: \'A DEFINITIONS ::= "
            "BEGIN Foo ::= SEQUENCE { a BOOLEAN, ..., [[ b BOOLEAN ]], [[ c "
            "BOOLEAN ]], ..., d BOOLEAN>!<, ... } END\': Expected \"}\".")

    def test_parse_x680_duplicated_enum_number_a_c_0(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.compile_string('A DEFINITIONS ::= BEGIN '
                                     'E ::= ENUMERATED { a, b, ..., c(0) } '
                                     'END')

        self.assertEqual(str(cm.exception),
                         "Duplicated ENUMERATED number 0 at line 1.")

    def test_parse_x680_duplicated_enum_number_c_d_2(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.compile_string('A DEFINITIONS ::= BEGIN '
                                     'E ::= ENUMERATED { a, b, ..., c, d(2) } '
                                     'END')

        self.assertEqual(str(cm.exception),
                         "Duplicated ENUMERATED number 2 at line 1.")

    def test_parse_x680_duplicated_enum_number_a_b_0(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.compile_string('A DEFINITIONS ::= BEGIN '
                                     'E ::= ENUMERATED { a(0), b(0) } '
                                     'END')

        self.assertEqual(str(cm.exception),
                         "Duplicated ENUMERATED number 0 at line 1.")


if __name__ == '__main__':
    unittest.main()



import unittest
from .utils import Asn1ToolsBaseTest
import asn1tools


class Asn1ToolsDerTest(Asn1ToolsBaseTest):

    maxDiff = None

    def test_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn', 'der')

        datas = [
            ('Boolean',                 True, b'\x01\x01\xff'),
            ('Boolean',                False, b'\x01\x01\x00'),
            ('Integer',                32768, b'\x02\x03\x00\x80\x00'),
            ('Integer',                32767, b'\x02\x02\x7f\xff'),
            ('Integer',                  256, b'\x02\x02\x01\x00'),
            ('Integer',                  255, b'\x02\x02\x00\xff'),
            ('Integer',                  128, b'\x02\x02\x00\x80'),
            ('Integer',                  127, b'\x02\x01\x7f'),
            ('Integer',                    1, b'\x02\x01\x01'),
            ('Integer',                    0, b'\x02\x01\x00'),
            ('Integer',                   -1, b'\x02\x01\xff'),
            ('Integer',                 -128, b'\x02\x01\x80'),
            ('Integer',                 -129, b'\x02\x02\xff\x7f'),
            ('Integer',                 -256, b'\x02\x02\xff\x00'),
            ('Integer',               -32768, b'\x02\x02\x80\x00'),
            ('Integer',               -32769, b'\x02\x03\xff\x7f\xff'),
            ('Bitstring',       (b'\x80', 1), b'\x03\x02\x07\x80'),
            ('Octetstring',          b'\x00', b'\x04\x01\x00'),
            ('Octetstring',    127 * b'\x55', b'\x04\x7f' + 127 * b'\x55'),
            ('Octetstring',    128 * b'\xaa', b'\x04\x81\x80' + 128 * b'\xaa'),
            ('Null',                    None, b'\x05\x00'),
            ('Objectidentifier',       '1.2', b'\x06\x01\x2a'),
            ('Enumerated',             'one', b'\x0a\x01\x01'),
            ('Utf8string',             'foo', b'\x0c\x03foo'),
            ('Sequence',                  {}, b'\x30\x00'),
            ('Sequence2',           {'a': 0}, b'\x30\x00'),
            ('Sequence2',           {'a': 1}, b'\x30\x03\x02\x01\x01'),
            ('Sequence13', {'a': [1]}, b'\x30\x05\xa0\x03\x02\x01\x01'),
            ('Sequence13', {'b': [1]}, b'\x30\x05\xa1\x03\x02\x01\x01'),
            ('Set',                       {}, b'\x31\x00'),
            ('Set2',                {'a': 1}, b'\x31\x00'),
            ('Set2',                {'a': 2}, b'\x31\x03\x02\x01\x02'),
            ('Numericstring',          '123', b'\x12\x03123'),
            ('Printablestring',        'foo', b'\x13\x03foo'),
            ('Ia5string',              'bar', b'\x16\x03bar'),
            ('Universalstring',        'bar', b'\x1c\x03bar'),
            ('Visiblestring',          'bar', b'\x1a\x03bar'),
            ('Generalstring',          'bar', b'\x1b\x03bar'),
            ('Bmpstring',             b'bar', b'\x1e\x03bar'),
            ('Teletexstring',         b'fum', b'\x14\x03fum'),
            ('Utctime',       '010203040506', b'\x17\x0d010203040506Z'),
            ('GeneralizedTime1',
             '20001231235959Z',
             b'\x18\x0f\x32\x30\x30\x30\x31\x32\x33\x31\x32\x33\x35\x39'
             b'\x35\x39\x5a'),
            ('SequenceOf',                [], b'0\x00'),
            ('SetOf',                     [], b'1\x00')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(all_types, type_name, decoded, encoded)

        with self.assertRaises(NotImplementedError):
            all_types.encode('Sequence12', {'a': [{'a': []}]})

        # ToDo: Should return {'a': [{'a': []}]}
        self.assertEqual(
            all_types.decode('Sequence12', b'\x30\x04\xa0\x02\x30\x00'),
            {})

    def test_repr_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn',
                                            'der')

        self.assertEqual(repr(all_types.types['Boolean']), 'Boolean(Boolean)')
        self.assertEqual(repr(all_types.types['Integer']), 'Integer(Integer)')
        self.assertEqual(repr(all_types.types['Bitstring']), 'BitString(Bitstring)')
        self.assertEqual(repr(all_types.types['Octetstring']), 'OctetString(Octetstring)')
        self.assertEqual(repr(all_types.types['Null']), 'Null(Null)')
        self.assertEqual(repr(all_types.types['Objectidentifier']),
                         'ObjectIdentifier(Objectidentifier)')
        self.assertEqual(repr(all_types.types['Enumerated']), 'Enumerated(Enumerated)')
        self.assertEqual(repr(all_types.types['Utf8string']), 'UTF8String(Utf8string)')
        self.assertEqual(repr(all_types.types['Sequence']), 'Sequence(Sequence, [])')
        self.assertEqual(repr(all_types.types['Set']), 'Set(Set, [])')
        self.assertEqual(repr(all_types.types['Sequence2']),
                         'Sequence(Sequence2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Set2']), 'Set(Set2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Numericstring']),
                         'NumericString(Numericstring)')
        self.assertEqual(repr(all_types.types['Printablestring']),
                         'PrintableString(Printablestring)')
        self.assertEqual(repr(all_types.types['Ia5string']), 'IA5String(Ia5string)')
        self.assertEqual(repr(all_types.types['Universalstring']),
                         'UniversalString(Universalstring)')
        self.assertEqual(repr(all_types.types['Visiblestring']),
                         'VisibleString(Visiblestring)')
        self.assertEqual(repr(all_types.types['Generalstring']),
                         'GeneralString(Generalstring)')
        self.assertEqual(repr(all_types.types['Bmpstring']),
                         'BMPString(Bmpstring)')
        self.assertEqual(repr(all_types.types['Teletexstring']),
                         'TeletexString(Teletexstring)')
        self.assertEqual(repr(all_types.types['Utctime']), 'UTCTime(Utctime)')
        self.assertEqual(repr(all_types.types['GeneralizedTime1']),
                         'GeneralizedTime(GeneralizedTime1)')
        self.assertEqual(repr(all_types.types['SequenceOf']),
                         'SequenceOf(SequenceOf, Integer())')
        self.assertEqual(repr(all_types.types['SetOf']), 'SetOf(SetOf, Integer())')

    def test_all_types_automatic_tags(self):
        all_types = asn1tools.compile_files(
            'tests/files/all_types_automatic_tags.asn', 'der')

        datas = [
            ('Sequence3',
             {'a': 1, 'c': 2,'d': True},
             b'\x30\x09\x80\x01\x01\x82\x01\x02\x83\x01\xff')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(all_types, type_name, decoded, encoded)

    def test_decode_length(self):
        foo = asn1tools.compile_files('tests/files/foo.asn', 'der')

        # The length can be decoded.
        datas = [
            (b'0\x0e\x02\x01\x01\x16\x09Is 1+1=3?',   16),
            (b'0\x10\x02\x02\x01\x16\x09Is 1+10=14?', 18),
            (b'0\x0d',                                15),
            (b'0\x84\x00\x00\x00\xb8',                190)
        ]

        for encoded_message, decoded_length in datas:
            length = foo.decode_length(encoded_message)
            self.assertEqual(length, decoded_length)

        # The length cannot be decoded.
        datas = [
            b'0',
            b'',
            b'0\x84\x00\x00\x00'
        ]

        for encoded in datas:
            self.assertIsNone(foo.decode_length(encoded))

    def test_long_tag(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS IMPLICIT TAGS ::= BEGIN "
            "A ::= [31] INTEGER "
            "B ::= [500] INTEGER "
            "END")

        datas = [
            ('A', 1, b'\x9f\x1f\x01\x01'),
            ('B', 1, b'\x9f\x83\x74\x01\x01')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

    def test_graphic_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= GraphicString "
            "END")

        datas = [
            ('A', 'f', b'\x19\x01\x66')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)


if __name__ == '__main__':
    unittest.main()



EXPECTED = {'ITS-Container': {'extensibility-implied': False,
                   'imports': {},
                   'object-classes': {},
                   'object-sets': {},
                   'tags': 'AUTOMATIC',
                   'types': {'AccelerationConfidence': {'restricted-to': [(0,
                                                                           102)],
                                                        'type': 'INTEGER'},
                             'AccelerationControl': {'size': [7],
                                                     'type': 'BIT STRING'},
                             'AccidentSubCauseCode': {'restricted-to': [(0,
                                                                         255)],
                                                      'type': 'INTEGER'},
                             'ActionID': {'members': [{'name': 'originatingStationID',
                                                       'type': 'StationID'},
                                                      {'name': 'sequenceNumber',
                                                       'type': 'SequenceNumber'}],
                                          'type': 'SEQUENCE'},
                             'AdverseWeatherCondition-AdhesionSubCauseCode': {'restricted-to': [(0,
                                                                                                 255)],
                                                                              'type': 'INTEGER'},
                             'AdverseWeatherCondition-ExtremeWeatherConditionSubCauseCode': {'restricted-to': [(0,
                                                                                                                255)],
                                                                                             'type': 'INTEGER'},
                             'AdverseWeatherCondition-PrecipitationSubCauseCode': {'restricted-to': [(0,
                                                                                                      255)],
                                                                                   'type': 'INTEGER'},
                             'AdverseWeatherCondition-VisibilitySubCauseCode': {'restricted-to': [(0,
                                                                                                   255)],
                                                                                'type': 'INTEGER'},
                             'Altitude': {'members': [{'name': 'altitudeValue',
                                                       'type': 'AltitudeValue'},
                                                      {'name': 'altitudeConfidence',
                                                       'type': 'AltitudeConfidence'}],
                                          'type': 'SEQUENCE'},
                             'AltitudeConfidence': {'type': 'ENUMERATED',
                                                    'values': [('alt-000-01',
                                                                0),
                                                               ('alt-000-02',
                                                                1),
                                                               ('alt-000-05',
                                                                2),
                                                               ('alt-000-10',
                                                                3),
                                                               ('alt-000-20',
                                                                4),
                                                               ('alt-000-50',
                                                                5),
                                                               ('alt-001-00',
                                                                6),
                                                               ('alt-002-00',
                                                                7),
                                                               ('alt-005-00',
                                                                8),
                                                               ('alt-010-00',
                                                                9),
                                                               ('alt-020-00',
                                                                10),
                                                               ('alt-050-00',
                                                                11),
                                                               ('alt-100-00',
                                                                12),
                                                               ('alt-200-00',
                                                                13),
                                                               ('outOfRange',
                                                                14),
                                                               ('unavailable',
                                                                15)]},
                             'AltitudeValue': {'restricted-to': [(-100000,
                                                                  800001)],
                                               'type': 'INTEGER'},
                             'CauseCode': {'members': [{'name': 'causeCode',
                                                        'type': 'CauseCodeType'},
                                                       {'name': 'subCauseCode',
                                                        'type': 'SubCauseCodeType'}],
                                           'type': 'SEQUENCE'},
                             'CauseCodeType': {'restricted-to': [(0, 255)],
                                               'type': 'INTEGER'},
                             'CenDsrcTollingZone': {'members': [{'name': 'protectedZoneLatitude',
                                                                 'type': 'Latitude'},
                                                                {'name': 'protectedZoneLongitude',
                                                                 'type': 'Longitude'},
                                                                {'name': 'cenDsrcTollingZoneID',
                                                                 'optional': True,
                                                                 'type': 'CenDsrcTollingZoneID'}],
                                                    'type': 'SEQUENCE'},
                             'CenDsrcTollingZoneID': {'type': 'ProtectedZoneID'},
                             'ClosedLanes': {'members': [{'name': 'hardShoulderStatus',
                                                          'optional': True,
                                                          'type': 'HardShoulderStatus'},
                                                         {'name': 'drivingLaneStatus',
                                                          'type': 'DrivingLaneStatus'},
                                                         '...'],
                                             'type': 'SEQUENCE'},
                             'CollisionRiskSubCauseCode': {'restricted-to': [(0,
                                                                              255)],
                                                           'type': 'INTEGER'},
                             'Curvature': {'members': [{'name': 'curvatureValue',
                                                        'type': 'CurvatureValue'},
                                                       {'name': 'curvatureConfidence',
                                                        'type': 'CurvatureConfidence'}],
                                           'type': 'SEQUENCE'},
                             'CurvatureCalculationMode': {'type': 'ENUMERATED',
                                                          'values': [('yawRateUsed',
                                                                      0),
                                                                     ('yawRateNotUsed',
                                                                      1),
                                                                     ('unavailable',
                                                                      2),
                                                                     '...']},
                             'CurvatureConfidence': {'type': 'ENUMERATED',
                                                     'values': [('onePerMeter-0-00002',
                                                                 0),
                                                                ('onePerMeter-0-0001',
                                                                 1),
                                                                ('onePerMeter-0-0005',
                                                                 2),
                                                                ('onePerMeter-0-002',
                                                                 3),
                                                                ('onePerMeter-0-01',
                                                                 4),
                                                                ('onePerMeter-0-1',
                                                                 5),
                                                                ('outOfRange',
                                                                 6),
                                                                ('unavailable',
                                                                 7)]},
                             'CurvatureValue': {'restricted-to': [(-30000,
                                                                   30001)],
                                                'type': 'INTEGER'},
                             'DangerousEndOfQueueSubCauseCode': {'restricted-to': [(0,
                                                                                    255)],
                                                                 'type': 'INTEGER'},
                             'DangerousGoodsBasic': {'type': 'ENUMERATED',
                                                     'values': [('explosives1',
                                                                 0),
                                                                ('explosives2',
                                                                 1),
                                                                ('explosives3',
                                                                 2),
                                                                ('explosives4',
                                                                 3),
                                                                ('explosives5',
                                                                 4),
                                                                ('explosives6',
                                                                 5),
                                                                ('flammableGases',
                                                                 6),
                                                                ('nonFlammableGases',
                                                                 7),
                                                                ('toxicGases',
                                                                 8),
                                                                ('flammableLiquids',
                                                                 9),
                                                                ('flammableSolids',
                                                                 10),
                                                                ('substancesLiableToSpontaneousCombustion',
                                                                 11),
                                                                ('substancesEmittingFlammableGasesUponContactWithWater',
                                                                 12),
                                                                ('oxidizingSubstances',
                                                                 13),
                                                                ('organicPeroxides',
                                                                 14),
                                                                ('toxicSubstances',
                                                                 15),
                                                                ('infectiousSubstances',
                                                                 16),
                                                                ('radioactiveMaterial',
                                                                 17),
                                                                ('corrosiveSubstances',
                                                                 18),
                                                                ('miscellaneousDangerousSubstances',
                                                                 19)]},
                             'DangerousGoodsExtended': {'members': [{'name': 'dangerousGoodsType',
                                                                     'type': 'DangerousGoodsBasic'},
                                                                    {'name': 'unNumber',
                                                                     'restricted-to': [(0,
                                                                                        9999)],
                                                                     'type': 'INTEGER'},
                                                                    {'name': 'elevatedTemperature',
                                                                     'type': 'BOOLEAN'},
                                                                    {'name': 'tunnelsRestricted',
                                                                     'type': 'BOOLEAN'},
                                                                    {'name': 'limitedQuantity',
                                                                     'type': 'BOOLEAN'},
                                                                    {'name': 'emergencyActionCode',
                                                                     'optional': True,
                                                                     'size': [(1,
                                                                               24)],
                                                                     'type': 'IA5String'},
                                                                    {'name': 'phoneNumber',
                                                                     'optional': True,
                                                                     'size': [(1,
                                                                               24)],
                                                                     'type': 'IA5String'},
                                                                    {'name': 'companyName',
                                                                     'optional': True,
                                                                     'size': [(1,
                                                                               24)],
                                                                     'type': 'UTF8String'}],
                                                        'type': 'SEQUENCE'},
                             'DangerousSituationSubCauseCode': {'restricted-to': [(0,
                                                                                   255)],
                                                                'type': 'INTEGER'},
                             'DeltaAltitude': {'restricted-to': [(-12700,
                                                                  12800)],
                                               'type': 'INTEGER'},
                             'DeltaLatitude': {'restricted-to': [(-131071,
                                                                  131072)],
                                               'type': 'INTEGER'},
                             'DeltaLongitude': {'restricted-to': [(-131071,
                                                                   131072)],
                                                'type': 'INTEGER'},
                             'DeltaReferencePosition': {'members': [{'name': 'deltaLatitude',
                                                                     'type': 'DeltaLatitude'},
                                                                    {'name': 'deltaLongitude',
                                                                     'type': 'DeltaLongitude'},
                                                                    {'name': 'deltaAltitude',
                                                                     'type': 'DeltaAltitude'}],
                                                        'type': 'SEQUENCE'},
                             'DriveDirection': {'type': 'ENUMERATED',
                                                'values': [('forward', 0),
                                                           ('backward', 1),
                                                           ('unavailable', 2)]},
                             'DrivingLaneStatus': {'size': [(1, 14)],
                                                   'type': 'BIT STRING'},
                             'EmbarkationStatus': {'type': 'BOOLEAN'},
                             'EmergencyPriority': {'size': [2],
                                                   'type': 'BIT STRING'},
                             'EmergencyVehicleApproachingSubCauseCode': {'restricted-to': [(0,
                                                                                            255)],
                                                                         'type': 'INTEGER'},
                             'EnergyStorageType': {'size': [7],
                                                   'type': 'BIT STRING'},
                             'EventHistory': {'element': {'type': 'EventPoint'},
                                              'size': [(1, 23)],
                                              'type': 'SEQUENCE OF'},
                             'EventPoint': {'members': [{'name': 'eventPosition',
                                                         'type': 'DeltaReferencePosition'},
                                                        {'name': 'eventDeltaTime',
                                                         'optional': True,
                                                         'type': 'PathDeltaTime'},
                                                        {'name': 'informationQuality',
                                                         'type': 'InformationQuality'}],
                                            'type': 'SEQUENCE'},
                             'ExteriorLights': {'size': [8],
                                                'type': 'BIT STRING'},
                             'HardShoulderStatus': {'type': 'ENUMERATED',
                                                    'values': [('availableForStopping',
                                                                0),
                                                               ('closed', 1),
                                                               ('availableForDriving',
                                                                2)]},
                             'HazardousLocation-AnimalOnTheRoadSubCauseCode': {'restricted-to': [(0,
                                                                                                  255)],
                                                                               'type': 'INTEGER'},
                             'HazardousLocation-DangerousCurveSubCauseCode': {'restricted-to': [(0,
                                                                                                 255)],
                                                                              'type': 'INTEGER'},
                             'HazardousLocation-ObstacleOnTheRoadSubCauseCode': {'restricted-to': [(0,
                                                                                                    255)],
                                                                                 'type': 'INTEGER'},
                             'HazardousLocation-SurfaceConditionSubCauseCode': {'restricted-to': [(0,
                                                                                                   255)],
                                                                                'type': 'INTEGER'},
                             'Heading': {'members': [{'name': 'headingValue',
                                                      'type': 'HeadingValue'},
                                                     {'name': 'headingConfidence',
                                                      'type': 'HeadingConfidence'}],
                                         'type': 'SEQUENCE'},
                             'HeadingConfidence': {'restricted-to': [(1, 127)],
                                                   'type': 'INTEGER'},
                             'HeadingValue': {'restricted-to': [(0, 3601)],
                                              'type': 'INTEGER'},
                             'HeightLonCarr': {'restricted-to': [(1, 100)],
                                               'type': 'INTEGER'},
                             'HumanPresenceOnTheRoadSubCauseCode': {'restricted-to': [(0,
                                                                                       255)],
                                                                    'type': 'INTEGER'},
                             'HumanProblemSubCauseCode': {'restricted-to': [(0,
                                                                             255)],
                                                          'type': 'INTEGER'},
                             'InformationQuality': {'restricted-to': [(0, 7)],
                                                    'type': 'INTEGER'},
                             'ItineraryPath': {'element': {'type': 'ReferencePosition'},
                                               'size': [(1, 40)],
                                               'type': 'SEQUENCE OF'},
                             'ItsPduHeader': {'members': [{'name': 'protocolVersion',
                                                           'restricted-to': [(0,
                                                                              255)],
                                                           'type': 'INTEGER'},
                                                          {'name': 'messageID',
                                                           'restricted-to': [(0,
                                                                              255)],
                                                           'type': 'INTEGER'},
                                                          {'name': 'stationID',
                                                           'type': 'StationID'}],
                                              'type': 'SEQUENCE'},
                             'LanePosition': {'restricted-to': [(-1, 14)],
                                              'type': 'INTEGER'},
                             'LateralAcceleration': {'members': [{'name': 'lateralAccelerationValue',
                                                                  'type': 'LateralAccelerationValue'},
                                                                 {'name': 'lateralAccelerationConfidence',
                                                                  'type': 'AccelerationConfidence'}],
                                                     'type': 'SEQUENCE'},
                             'LateralAccelerationValue': {'restricted-to': [(-160,
                                                                             161)],
                                                          'type': 'INTEGER'},
                             'Latitude': {'restricted-to': [(-900000000,
                                                             900000001)],
                                          'type': 'INTEGER'},
                             'LightBarSirenInUse': {'size': [2],
                                                    'type': 'BIT STRING'},
                             'Longitude': {'restricted-to': [(-1800000000,
                                                              1800000001)],
                                           'type': 'INTEGER'},
                             'LongitudinalAcceleration': {'members': [{'name': 'longitudinalAccelerationValue',
                                                                       'type': 'LongitudinalAccelerationValue'},
                                                                      {'name': 'longitudinalAccelerationConfidence',
                                                                       'type': 'AccelerationConfidence'}],
                                                          'type': 'SEQUENCE'},
                             'LongitudinalAccelerationValue': {'restricted-to': [(-160,
                                                                                  161)],
                                                               'type': 'INTEGER'},
                             'NumberOfOccupants': {'restricted-to': [(0, 127)],
                                                   'type': 'INTEGER'},
                             'PathDeltaTime': {'restricted-to': [(1, 65535),
                                                                 '...'],
                                               'type': 'INTEGER'},
                             'PathHistory': {'element': {'type': 'PathPoint'},
                                             'size': [(0, 40)],
                                             'type': 'SEQUENCE OF'},
                             'PathPoint': {'members': [{'name': 'pathPosition',
                                                        'type': 'DeltaReferencePosition'},
                                                       {'name': 'pathDeltaTime',
                                                        'optional': True,
                                                        'type': 'PathDeltaTime'}],
                                           'type': 'SEQUENCE'},
                             'PerformanceClass': {'restricted-to': [(0, 7)],
                                                  'type': 'INTEGER'},
                             'PosCentMass': {'restricted-to': [(1, 63)],
                                             'type': 'INTEGER'},
                             'PosConfidenceEllipse': {'members': [{'name': 'semiMajorConfidence',
                                                                   'type': 'SemiAxisLength'},
                                                                  {'name': 'semiMinorConfidence',
                                                                   'type': 'SemiAxisLength'},
                                                                  {'name': 'semiMajorOrientation',
                                                                   'type': 'HeadingValue'}],
                                                      'type': 'SEQUENCE'},
                             'PosFrontAx': {'restricted-to': [(1, 20)],
                                            'type': 'INTEGER'},
                             'PosLonCarr': {'restricted-to': [(1, 127)],
                                            'type': 'INTEGER'},
                             'PosPillar': {'restricted-to': [(1, 30)],
                                           'type': 'INTEGER'},
                             'PositionOfOccupants': {'size': [20],
                                                     'type': 'BIT STRING'},
                             'PositionOfPillars': {'element': {'type': 'PosPillar'},
                                                   'size': [(1, 3), '...'],
                                                   'type': 'SEQUENCE OF'},
                             'PositioningSolutionType': {'type': 'ENUMERATED',
                                                         'values': [('noPositioningSolution',
                                                                     0),
                                                                    ('sGNSS',
                                                                     1),
                                                                    ('dGNSS',
                                                                     2),
                                                                    ('sGNSSplusDR',
                                                                     3),
                                                                    ('dGNSSplusDR',
                                                                     4),
                                                                    ('dR', 5),
                                                                    '...']},
                             'PostCrashSubCauseCode': {'restricted-to': [(0,
                                                                          255)],
                                                       'type': 'INTEGER'},
                             'ProtectedCommunicationZone': {'members': [{'name': 'protectedZoneType',
                                                                         'type': 'ProtectedZoneType'},
                                                                        {'name': 'expiryTime',
                                                                         'optional': True,
                                                                         'type': 'TimestampIts'},
                                                                        {'name': 'protectedZoneLatitude',
                                                                         'type': 'Latitude'},
                                                                        {'name': 'protectedZoneLongitude',
                                                                         'type': 'Longitude'},
                                                                        {'name': 'protectedZoneRadius',
                                                                         'optional': True,
                                                                         'type': 'ProtectedZoneRadius'},
                                                                        {'name': 'protectedZoneID',
                                                                         'optional': True,
                                                                         'type': 'ProtectedZoneID'}],
                                                            'type': 'SEQUENCE'},
                             'ProtectedCommunicationZonesRSU': {'element': {'type': 'ProtectedCommunicationZone'},
                                                                'size': [(1,
                                                                          16)],
                                                                'type': 'SEQUENCE '
                                                                        'OF'},
                             'ProtectedZoneID': {'restricted-to': [(0,
                                                                    134217727)],
                                                 'type': 'INTEGER'},
                             'ProtectedZoneRadius': {'restricted-to': [(1, 255),
                                                                       '...'],
                                                     'type': 'INTEGER'},
                             'ProtectedZoneType': {'type': 'ENUMERATED',
                                                   'values': [('cenDsrcTolling',
                                                               0),
                                                              '...']},
                             'PtActivation': {'members': [{'name': 'ptActivationType',
                                                           'type': 'PtActivationType'},
                                                          {'name': 'ptActivationData',
                                                           'type': 'PtActivationData'}],
                                              'type': 'SEQUENCE'},
                             'PtActivationData': {'size': [(1, 20)],
                                                  'type': 'OCTET STRING'},
                             'PtActivationType': {'restricted-to': [(0, 255)],
                                                  'type': 'INTEGER'},
                             'ReferencePosition': {'members': [{'name': 'latitude',
                                                                'type': 'Latitude'},
                                                               {'name': 'longitude',
                                                                'type': 'Longitude'},
                                                               {'name': 'positionConfidenceEllipse',
                                                                'type': 'PosConfidenceEllipse'},
                                                               {'name': 'altitude',
                                                                'type': 'Altitude'}],
                                                   'type': 'SEQUENCE'},
                             'RelevanceDistance': {'type': 'ENUMERATED',
                                                   'values': [('lessThan50m',
                                                               0),
                                                              ('lessThan100m',
                                                               1),
                                                              ('lessThan200m',
                                                               2),
                                                              ('lessThan500m',
                                                               3),
                                                              ('lessThan1000m',
                                                               4),
                                                              ('lessThan5km',
                                                               5),
                                                              ('lessThan10km',
                                                               6),
                                                              ('over10km', 7)]},
                             'RelevanceTrafficDirection': {'type': 'ENUMERATED',
                                                           'values': [('allTrafficDirections',
                                                                       0),
                                                                      ('upstreamTraffic',
                                                                       1),
                                                                      ('downstreamTraffic',
                                                                       2),
                                                                      ('oppositeTraffic',
                                                                       3)]},
                             'RequestResponseIndication': {'type': 'ENUMERATED',
                                                           'values': [('request',
                                                                       0),
                                                                      ('response',
                                                                       1)]},
                             'RescueAndRecoveryWorkInProgressSubCauseCode': {'restricted-to': [(0,
                                                                                                255)],
                                                                             'type': 'INTEGER'},
                             'RestrictedTypes': {'element': {'type': 'StationType'},
                                                 'size': [(1, 3), '...'],
                                                 'type': 'SEQUENCE OF'},
                             'RoadType': {'type': 'ENUMERATED',
                                          'values': [('urban-NoStructuralSeparationToOppositeLanes',
                                                      0),
                                                     ('urban-WithStructuralSeparationToOppositeLanes',
                                                      1),
                                                     ('nonUrban-NoStructuralSeparationToOppositeLanes',
                                                      2),
                                                     ('nonUrban-WithStructuralSeparationToOppositeLanes',
                                                      3)]},
                             'RoadworksSubCauseCode': {'restricted-to': [(0,
                                                                          255)],
                                                       'type': 'INTEGER'},
                             'SemiAxisLength': {'restricted-to': [(0, 4095)],
                                                'type': 'INTEGER'},
                             'SequenceNumber': {'restricted-to': [(0, 65535)],
                                                'type': 'INTEGER'},
                             'SignalViolationSubCauseCode': {'restricted-to': [(0,
                                                                                255)],
                                                             'type': 'INTEGER'},
                             'SlowVehicleSubCauseCode': {'restricted-to': [(0,
                                                                            255)],
                                                         'type': 'INTEGER'},
                             'SpecialTransportType': {'size': [4],
                                                      'type': 'BIT STRING'},
                             'Speed': {'members': [{'name': 'speedValue',
                                                    'type': 'SpeedValue'},
                                                   {'name': 'speedConfidence',
                                                    'type': 'SpeedConfidence'}],
                                       'type': 'SEQUENCE'},
                             'SpeedConfidence': {'restricted-to': [(1, 127)],
                                                 'type': 'INTEGER'},
                             'SpeedLimit': {'restricted-to': [(1, 255)],
                                            'type': 'INTEGER'},
                             'SpeedValue': {'restricted-to': [(0, 16383)],
                                            'type': 'INTEGER'},
                             'StationID': {'restricted-to': [(0, 4294967295)],
                                           'type': 'INTEGER'},
                             'StationType': {'restricted-to': [(0, 255)],
                                             'type': 'INTEGER'},
                             'StationarySince': {'type': 'ENUMERATED',
                                                 'values': [('lessThan1Minute',
                                                             0),
                                                            ('lessThan2Minutes',
                                                             1),
                                                            ('lessThan15Minutes',
                                                             2),
                                                            ('equalOrGreater15Minutes',
                                                             3)]},
                             'StationaryVehicleSubCauseCode': {'restricted-to': [(0,
                                                                                  255)],
                                                               'type': 'INTEGER'},
                             'SteeringWheelAngle': {'members': [{'name': 'steeringWheelAngleValue',
                                                                 'type': 'SteeringWheelAngleValue'},
                                                                {'name': 'steeringWheelAngleConfidence',
                                                                 'type': 'SteeringWheelAngleConfidence'}],
                                                    'type': 'SEQUENCE'},
                             'SteeringWheelAngleConfidence': {'restricted-to': [(1,
                                                                                 127)],
                                                              'type': 'INTEGER'},
                             'SteeringWheelAngleValue': {'restricted-to': [(-511,
                                                                            512)],
                                                         'type': 'INTEGER'},
                             'SubCauseCodeType': {'restricted-to': [(0, 255)],
                                                  'type': 'INTEGER'},
                             'Temperature': {'restricted-to': [(-60, 67)],
                                             'type': 'INTEGER'},
                             'TimestampIts': {'restricted-to': [(0,
                                                                 4398046511103)],
                                              'type': 'INTEGER'},
                             'Traces': {'element': {'type': 'PathHistory'},
                                        'size': [(1, 7)],
                                        'type': 'SEQUENCE OF'},
                             'TrafficConditionSubCauseCode': {'restricted-to': [(0,
                                                                                 255)],
                                                              'type': 'INTEGER'},
                             'TrafficRule': {'type': 'ENUMERATED',
                                             'values': [('noPassing', 0),
                                                        ('noPassingForTrucks',
                                                         1),
                                                        ('passToRight', 2),
                                                        ('passToLeft', 3),
                                                        '...']},
                             'TransmissionInterval': {'restricted-to': [(1,
                                                                         10000)],
                                                      'type': 'INTEGER'},
                             'TurningRadius': {'restricted-to': [(1, 255)],
                                               'type': 'INTEGER'},
                             'VDS': {'size': [6], 'type': 'IA5String'},
                             'ValidityDuration': {'restricted-to': [(0, 86400)],
                                                  'type': 'INTEGER'},
                             'VehicleBreakdownSubCauseCode': {'restricted-to': [(0,
                                                                                 255)],
                                                              'type': 'INTEGER'},
                             'VehicleIdentification': {'members': [{'name': 'wMInumber',
                                                                    'optional': True,
                                                                    'type': 'WMInumber'},
                                                                   {'name': 'vDS',
                                                                    'optional': True,
                                                                    'type': 'VDS'},
                                                                   '...'],
                                                       'type': 'SEQUENCE'},
                             'VehicleLength': {'members': [{'name': 'vehicleLengthValue',
                                                            'type': 'VehicleLengthValue'},
                                                           {'name': 'vehicleLengthConfidenceIndication',
                                                            'type': 'VehicleLengthConfidenceIndication'}],
                                               'type': 'SEQUENCE'},
                             'VehicleLengthConfidenceIndication': {'type': 'ENUMERATED',
                                                                   'values': [('noTrailerPresent',
                                                                               0),
                                                                              ('trailerPresentWithKnownLength',
                                                                               1),
                                                                              ('trailerPresentWithUnknownLength',
                                                                               2),
                                                                              ('trailerPresenceIsUnknown',
                                                                               3),
                                                                              ('unavailable',
                                                                               4)]},
                             'VehicleLengthValue': {'restricted-to': [(1,
                                                                       1023)],
                                                    'type': 'INTEGER'},
                             'VehicleMass': {'restricted-to': [(1, 1024)],
                                             'type': 'INTEGER'},
                             'VehicleRole': {'type': 'ENUMERATED',
                                             'values': [('default', 0),
                                                        ('publicTransport', 1),
                                                        ('specialTransport', 2),
                                                        ('dangerousGoods', 3),
                                                        ('roadWork', 4),
                                                        ('rescue', 5),
                                                        ('emergency', 6),
                                                        ('safetyCar', 7),
                                                        ('agriculture', 8),
                                                        ('commercial', 9),
                                                        ('military', 10),
                                                        ('roadOperator', 11),
                                                        ('taxi', 12),
                                                        ('reserved1', 13),
                                                        ('reserved2', 14),
                                                        ('reserved3', 15)]},
                             'VehicleWidth': {'restricted-to': [(1, 62)],
                                              'type': 'INTEGER'},
                             'VerticalAcceleration': {'members': [{'name': 'verticalAccelerationValue',
                                                                   'type': 'VerticalAccelerationValue'},
                                                                  {'name': 'verticalAccelerationConfidence',
                                                                   'type': 'AccelerationConfidence'}],
                                                      'type': 'SEQUENCE'},
                             'VerticalAccelerationValue': {'restricted-to': [(-160,
                                                                              161)],
                                                           'type': 'INTEGER'},
                             'WMInumber': {'size': [(1, 3)],
                                           'type': 'IA5String'},
                             'WheelBaseVehicle': {'restricted-to': [(1, 127)],
                                                  'type': 'INTEGER'},
                             'WrongWayDrivingSubCauseCode': {'restricted-to': [(0,
                                                                                255)],
                                                             'type': 'INTEGER'},
                             'YawRate': {'members': [{'name': 'yawRateValue',
                                                      'type': 'YawRateValue'},
                                                     {'name': 'yawRateConfidence',
                                                      'type': 'YawRateConfidence'}],
                                         'type': 'SEQUENCE'},
                             'YawRateConfidence': {'type': 'ENUMERATED',
                                                   'values': [('degSec-000-01',
                                                               0),
                                                              ('degSec-000-05',
                                                               1),
                                                              ('degSec-000-10',
                                                               2),
                                                              ('degSec-001-00',
                                                               3),
                                                              ('degSec-005-00',
                                                               4),
                                                              ('degSec-010-00',
                                                               5),
                                                              ('degSec-100-00',
                                                               6),
                                                              ('outOfRange', 7),
                                                              ('unavailable',
                                                               8)]},
                             'YawRateValue': {'restricted-to': [(-32766,
                                                                 32767)],
                                              'type': 'INTEGER'}},
                   'values': {}}}


EXPECTED = {'CAM-PDU-Descriptions': {'extensibility-implied': False,
                          'imports': {'ITS-Container': ['ItsPduHeader',
                                                        'CauseCode',
                                                        'ReferencePosition',
                                                        'AccelerationControl',
                                                        'Curvature',
                                                        'CurvatureCalculationMode',
                                                        'Heading',
                                                        'LanePosition',
                                                        'EmergencyPriority',
                                                        'EmbarkationStatus',
                                                        'Speed',
                                                        'DriveDirection',
                                                        'LongitudinalAcceleration',
                                                        'LateralAcceleration',
                                                        'VerticalAcceleration',
                                                        'StationType',
                                                        'ExteriorLights',
                                                        'DangerousGoodsBasic',
                                                        'SpecialTransportType',
                                                        'LightBarSirenInUse',
                                                        'VehicleRole',
                                                        'VehicleLength',
                                                        'VehicleWidth',
                                                        'PathHistory',
                                                        'RoadworksSubCauseCode',
                                                        'ClosedLanes',
                                                        'TrafficRule',
                                                        'SpeedLimit',
                                                        'SteeringWheelAngle',
                                                        'PerformanceClass',
                                                        'YawRate',
                                                        'ProtectedCommunicationZone',
                                                        'PtActivation',
                                                        'Latitude',
                                                        'Longitude',
                                                        'ProtectedCommunicationZonesRSU',
                                                        'CenDsrcTollingZone']},
                          'object-classes': {},
                          'object-sets': {},
                          'tags': 'AUTOMATIC',
                          'types': {'BasicContainer': {'members': [{'name': 'stationType',
                                                                    'type': 'StationType'},
                                                                   {'name': 'referencePosition',
                                                                    'type': 'ReferencePosition'},
                                                                   '...'],
                                                       'type': 'SEQUENCE'},
                                    'BasicVehicleContainerHighFrequency': {'members': [{'name': 'heading',
                                                                                        'type': 'Heading'},
                                                                                       {'name': 'speed',
                                                                                        'type': 'Speed'},
                                                                                       {'name': 'driveDirection',
                                                                                        'type': 'DriveDirection'},
                                                                                       {'name': 'vehicleLength',
                                                                                        'type': 'VehicleLength'},
                                                                                       {'name': 'vehicleWidth',
                                                                                        'type': 'VehicleWidth'},
                                                                                       {'name': 'longitudinalAcceleration',
                                                                                        'type': 'LongitudinalAcceleration'},
                                                                                       {'name': 'curvature',
                                                                                        'type': 'Curvature'},
                                                                                       {'name': 'curvatureCalculationMode',
                                                                                        'type': 'CurvatureCalculationMode'},
                                                                                       {'name': 'yawRate',
                                                                                        'type': 'YawRate'},
                                                                                       {'name': 'accelerationControl',
                                                                                        'optional': True,
                                                                                        'type': 'AccelerationControl'},
                                                                                       {'name': 'lanePosition',
                                                                                        'optional': True,
                                                                                        'type': 'LanePosition'},
                                                                                       {'name': 'steeringWheelAngle',
                                                                                        'optional': True,
                                                                                        'type': 'SteeringWheelAngle'},
                                                                                       {'name': 'lateralAcceleration',
                                                                                        'optional': True,
                                                                                        'type': 'LateralAcceleration'},
                                                                                       {'name': 'verticalAcceleration',
                                                                                        'optional': True,
                                                                                        'type': 'VerticalAcceleration'},
                                                                                       {'name': 'performanceClass',
                                                                                        'optional': True,
                                                                                        'type': 'PerformanceClass'},
                                                                                       {'name': 'cenDsrcTollingZone',
                                                                                        'optional': True,
                                                                                        'type': 'CenDsrcTollingZone'}],
                                                                           'type': 'SEQUENCE'},
                                    'BasicVehicleContainerLowFrequency': {'members': [{'name': 'vehicleRole',
                                                                                       'type': 'VehicleRole'},
                                                                                      {'name': 'exteriorLights',
                                                                                       'type': 'ExteriorLights'},
                                                                                      {'name': 'pathHistory',
                                                                                       'type': 'PathHistory'}],
                                                                          'type': 'SEQUENCE'},
                                    'CAM': {'members': [{'name': 'header',
                                                         'type': 'ItsPduHeader'},
                                                        {'name': 'cam',
                                                         'type': 'CoopAwareness'}],
                                            'type': 'SEQUENCE'},
                                    'CamParameters': {'members': [{'name': 'basicContainer',
                                                                   'type': 'BasicContainer'},
                                                                  {'name': 'highFrequencyContainer',
                                                                   'type': 'HighFrequencyContainer'},
                                                                  {'name': 'lowFrequencyContainer',
                                                                   'optional': True,
                                                                   'type': 'LowFrequencyContainer'},
                                                                  {'name': 'specialVehicleContainer',
                                                                   'optional': True,
                                                                   'type': 'SpecialVehicleContainer'},
                                                                  '...'],
                                                      'type': 'SEQUENCE'},
                                    'CoopAwareness': {'members': [{'name': 'generationDeltaTime',
                                                                   'type': 'GenerationDeltaTime'},
                                                                  {'name': 'camParameters',
                                                                   'type': 'CamParameters'}],
                                                      'type': 'SEQUENCE'},
                                    'DangerousGoodsContainer': {'members': [{'name': 'dangerousGoodsBasic',
                                                                             'type': 'DangerousGoodsBasic'}],
                                                                'type': 'SEQUENCE'},
                                    'EmergencyContainer': {'members': [{'name': 'lightBarSirenInUse',
                                                                        'type': 'LightBarSirenInUse'},
                                                                       {'name': 'incidentIndication',
                                                                        'optional': True,
                                                                        'type': 'CauseCode'},
                                                                       {'name': 'emergencyPriority',
                                                                        'optional': True,
                                                                        'type': 'EmergencyPriority'}],
                                                           'type': 'SEQUENCE'},
                                    'GenerationDeltaTime': {'restricted-to': [(0,
                                                                               65535)],
                                                            'type': 'INTEGER'},
                                    'HighFrequencyContainer': {'members': [{'name': 'basicVehicleContainerHighFrequency',
                                                                            'type': 'BasicVehicleContainerHighFrequency'},
                                                                           {'name': 'rsuContainerHighFrequency',
                                                                            'type': 'RSUContainerHighFrequency'},
                                                                           '...'],
                                                               'type': 'CHOICE'},
                                    'LowFrequencyContainer': {'members': [{'name': 'basicVehicleContainerLowFrequency',
                                                                           'type': 'BasicVehicleContainerLowFrequency'},
                                                                          '...'],
                                                              'type': 'CHOICE'},
                                    'PublicTransportContainer': {'members': [{'name': 'embarkationStatus',
                                                                              'type': 'EmbarkationStatus'},
                                                                             {'name': 'ptActivation',
                                                                              'optional': True,
                                                                              'type': 'PtActivation'}],
                                                                 'type': 'SEQUENCE'},
                                    'RSUContainerHighFrequency': {'members': [{'name': 'protectedCommunicationZonesRSU',
                                                                               'optional': True,
                                                                               'type': 'ProtectedCommunicationZonesRSU'},
                                                                              '...'],
                                                                  'type': 'SEQUENCE'},
                                    'RescueContainer': {'members': [{'name': 'lightBarSirenInUse',
                                                                     'type': 'LightBarSirenInUse'}],
                                                        'type': 'SEQUENCE'},
                                    'RoadWorksContainerBasic': {'members': [{'name': 'roadworksSubCauseCode',
                                                                             'optional': True,
                                                                             'type': 'RoadworksSubCauseCode'},
                                                                            {'name': 'lightBarSirenInUse',
                                                                             'type': 'LightBarSirenInUse'},
                                                                            {'name': 'closedLanes',
                                                                             'optional': True,
                                                                             'type': 'ClosedLanes'}],
                                                                'type': 'SEQUENCE'},
                                    'SafetyCarContainer': {'members': [{'name': 'lightBarSirenInUse',
                                                                        'type': 'LightBarSirenInUse'},
                                                                       {'name': 'incidentIndication',
                                                                        'optional': True,
                                                                        'type': 'CauseCode'},
                                                                       {'name': 'trafficRule',
                                                                        'optional': True,
                                                                        'type': 'TrafficRule'},
                                                                       {'name': 'speedLimit',
                                                                        'optional': True,
                                                                        'type': 'SpeedLimit'}],
                                                           'type': 'SEQUENCE'},
                                    'SpecialTransportContainer': {'members': [{'name': 'specialTransportType',
                                                                               'type': 'SpecialTransportType'},
                                                                              {'name': 'lightBarSirenInUse',
                                                                               'type': 'LightBarSirenInUse'}],
                                                                  'type': 'SEQUENCE'},
                                    'SpecialVehicleContainer': {'members': [{'name': 'publicTransportContainer',
                                                                             'type': 'PublicTransportContainer'},
                                                                            {'name': 'specialTransportContainer',
                                                                             'type': 'SpecialTransportContainer'},
                                                                            {'name': 'dangerousGoodsContainer',
                                                                             'type': 'DangerousGoodsContainer'},
                                                                            {'name': 'roadWorksContainerBasic',
                                                                             'type': 'RoadWorksContainerBasic'},
                                                                            {'name': 'rescueContainer',
                                                                             'type': 'RescueContainer'},
                                                                            {'name': 'emergencyContainer',
                                                                             'type': 'EmergencyContainer'},
                                                                            {'name': 'safetyCarContainer',
                                                                             'type': 'SafetyCarContainer'},
                                                                            '...'],
                                                                'type': 'CHOICE'}},
                          'values': {}}}



"""Packed Encoding Rules (PER) codec.

"""

from . import EncodeError
from . import DecodeError
from . import compiler
from .compiler import enum_values_as_dict


class DecodeChoiceError(Exception):
    pass


class Encoder(object):

    def __init__(self):
        self.byte = 0
        self.index = 7
        self.buf = bytearray()

    def append_bit(self, bit):
        """Append given bit.

        """

        self.byte |= (bit << self.index)
        self.index -= 1

        if self.index == -1:
            self.buf.append(self.byte)
            self.byte = 0
            self.index = 7

    def append_bits(self, data, number_of_bits):
        """Append given bits.

        """

        for i in range(number_of_bits):
            self.append_bit((bytearray(data)[i // 8] >> (7 - (i % 8))) & 0x1)

    def append_integer(self, value, number_of_bits):
        """Append given integer value.

        """

        for i in range(number_of_bits):
            self.append_bit((value >> (number_of_bits - i - 1)) & 0x1)

    def append_bytes(self, data):
        """Append given data aligned to a byte boundary.

        """

        if self.index != 7:
            self.buf.append(self.byte)
            self.byte = 0
            self.index = 7

        self.buf.extend(data)

    def as_bytearray(self):
        """Return the bits as a bytearray.

        """

        if self.index < 7:
            return self.buf + bytearray([self.byte])
        else:
            return self.buf

    def append_length_determinant(self, length):
        if length < 128:
            encoded = bytearray([length])
        elif length < 16384:
            encoded = bytearray([(0x80 | (length >> 8)), (length & 0xff)])
        else:
            raise NotImplementedError(
                'Length determinant >=16384 is not yet supported.')

        self.append_bytes(encoded)

    def append_unconstrained_whole_number(self, value):
        number_of_bits = value.bit_length()

        if value < 0:
            number_of_bytes = ((number_of_bits + 7) // 8)
            value = ((1 << (8 * number_of_bytes)) + value)

            if (value & (1 << (8 * number_of_bytes - 1))) == 0:
                value |= (0xff << (8 * number_of_bytes))
                number_of_bytes += 1
        elif value > 0:
            number_of_bytes = ((number_of_bits + 7) // 8)

            if number_of_bits == (8 * number_of_bytes):
                number_of_bytes += 1
        else:
            number_of_bytes = 1

        self.append_length_determinant(number_of_bytes)
        self.append_integer(value, 8 * number_of_bytes)

    def __repr__(self):
        return str(self.as_bytearray())


class Decoder(object):

    def __init__(self, encoded):
        self.byte = None
        self.index = -1
        self.offset = 0
        self.buf = encoded

    def read_bit(self):
        """Read a bit.

        """

        if self.index == -1:
            self.byte = self.buf[self.offset]
            self.index = 7
            self.offset += 1

        bit = ((self.byte >> self.index) & 0x1)
        self.index -= 1

        return bit

    def read_integer(self, number_of_bits):
        """Read an integer value of given number of bits.

        """

        value = 0

        for _ in range(number_of_bits):
            value <<= 1
            value |= self.read_bit()

        return value

    def read_bytes(self, number_of_bytes):
        """Read given number of bytes.

        """

        self.index = -1
        data = self.buf[self.offset:self.offset + number_of_bytes]
        self.offset += number_of_bytes

        return data

    def read_length_determinant(self):
        value = self.read_bytes(1)[0]

        if (value & 0x80) == 0x00:
            return value
        elif (value & 0xc0) == 0x80:
            return (((value & 0x7f) << 8) | (self.read_integer(8)))
        else:
            raise NotImplementedError(
                'Length determinant type 0x{:02x} not yet supported.'.format(
                    value))

    def read_unconstrained_whole_number(self, number_of_bytes):
        decoded = self.read_integer(8 * number_of_bytes)
        number_of_bits = (8 * number_of_bytes)

        if decoded & (1 << (number_of_bits - 1)):
            mask = ((1 << number_of_bits) - 1)
            decoded = (decoded - mask)
            decoded -= 1

        return decoded


def size_as_number_of_bits(size):
    """Returns the minimum number of bits needed to fit given positive
    integer.

    """

    if size == 0:
        return 0
    else:
        return len('{:b}'.format(size))


class Type(object):

    def __init__(self, name, type_name):
        self.name = name
        self.type_name = type_name
        self.optional = False
        self.default = None
        self.tag = None

    def set_size_range(self, minimum, maximum):
        pass


class Integer(Type):

    def __init__(self, name, minimum, maximum, has_extension_marker):
        super(Integer, self).__init__(name, 'INTEGER')
        self.minimum = minimum
        self.maximum = maximum
        self.has_extension_marker = has_extension_marker

        if minimum is None or maximum is None:
            self.number_of_bits = None
        else:
            size = self.maximum - self.minimum
            self.number_of_bits = size_as_number_of_bits(size)

    def encode(self, data, encoder):
        if self.has_extension_marker:
            encoder.append_bit(0)

        if self.number_of_bits is None:
            encoder.append_unconstrained_whole_number(data)
        else:
            encoder.append_integer(data - self.minimum, self.number_of_bits)

    def decode(self, decoder):
        if self.has_extension_marker:
            bit = decoder.read_bit()

            if bit:
                raise NotImplementedError('Extension is not implemented.')

        if self.number_of_bits is None:
            length = decoder.read_length_determinant()
            value = decoder.read_unconstrained_whole_number(length)
        else:
            value = decoder.read_integer(self.number_of_bits)
            value += self.minimum

        return value

    def __repr__(self):
        return 'Integer({})'.format(self.name)


class Real(Type):

    def __init__(self, name):
        super(Real, self).__init__(name, 'REAL')

    def encode(self, data, encoder):
        raise NotImplementedError()

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'Real({})'.format(self.name)


class Boolean(Type):

    def __init__(self, name):
        super(Boolean, self).__init__(name, 'BOOLEAN')

    def encode(self, data, encoder):
        encoder.append_bit(bool(data))

    def decode(self, decoder):
        return bool(decoder.read_bit())

    def __repr__(self):
        return 'Boolean({})'.format(self.name)


class IA5String(Type):

    def __init__(self, name):
        super(IA5String, self).__init__(name, 'IA5String')

    def encode(self, data, encoder):
        encoder.append_bytes(bytearray([len(data)]) + data.encode('ascii'))

    def decode(self, decoder):
        length = decoder.read_bytes(1)[0]

        return decoder.read_bytes(length).decode('ascii')

    def __repr__(self):
        return 'IA5String({})'.format(self.name)


class NumericString(Type):

    def __init__(self, name):
        super(NumericString, self).__init__(name, 'NumericString')

    def encode(self, data, encoder):
        raise NotImplementedError()

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'NumericString({})'.format(self.name)


class Sequence(Type):

    def __init__(self, name, members):
        super(Sequence, self).__init__(name, 'SEQUENCE')
        self.members = members
        self.optionals = [member.name
                          for member in members
                          if member.optional]

    def encode(self, data, encoder):
        for optional in self.optionals:
            encoder.append_bit(optional in data)

        for member in self.members:
            name = member.name

            if name in data:
                member.encode(data[name], encoder)
            elif member.optional:
                pass
            elif member.default is not None:
                member.encode(member.default, encoder)
            else:
                raise EncodeError(
                    "Sequence member '{}' not found in {}.".format(
                        name,
                        data))

    def decode(self, decoder):
        values = {}

        optionals = {optional: decoder.read_bit()
                     for optional in self.optionals}

        for member in self.members:
            if not optionals.get(member.name, True):
                continue

            try:
                value = member.decode(decoder)
            except (DecodeError, IndexError) as e:
                if member.optional:
                    continue

                if member.default is None:
                    if isinstance(e, IndexError):
                        e = DecodeError('out of data at offset {}'.format(-1))

                    e.location.append(member.name)
                    raise e

                value = member.default

            values[member.name] = value

        return values

    def __repr__(self):
        return 'Sequence({}, [{}])'.format(
            self.name,
            ', '.join([repr(member) for member in self.members]))


class Set(Type):

    def __init__(self, name, members):
        super(Set, self).__init__(name, 'SET')
        self.members = members

    def encode(self, data, encoder):
        raise NotImplementedError()

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'Set({}, [{}])'.format(
            self.name,
            ', '.join([repr(member) for member in self.members]))


class SequenceOf(Type):

    def __init__(self, name, element_type):
        super(SequenceOf, self).__init__(name, 'SEQUENCE OF')
        self.element_type = element_type

    def encode(self, data, encoder):
        encoder.append_bytes(bytearray([len(data)]))

        for entry in data:
            self.element_type.encode(entry, encoder)

    def decode(self, decoder):
        number_of_elements = decoder.read_bytes(1)[0]
        decoded = []

        for _ in range(number_of_elements):
            decoded_element = self.element_type.decode(decoder)
            decoded.append(decoded_element)

        return decoded

    def __repr__(self):
        return 'SequenceOf({}, {})'.format(self.name,
                                           self.element_type)


class SetOf(Type):

    def __init__(self, name, element_type):
        super(SetOf, self).__init__(name, 'SET OF')
        self.element_type = element_type

    def encode(self, data, encoder):
        raise NotImplementedError()

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'SetOf({}, {})'.format(self.name,
                                      self.element_type)


class BitString(Type):

    def __init__(self, name, size):
        super(BitString, self).__init__(name, 'BIT STRING')
        self.size = size

    def encode(self, data, encoder):
        if self.size is None:
            encoder.append_bytes(bytearray([data[1]]) + data[0])
        else:
            encoder.append_bits(*data)

    def decode(self, decoder):
        number_of_bits = decoder.read_bytes(1)[0]

        return (decoder.read_bytes((number_of_bits + 7) // 8), number_of_bits)

    def __repr__(self):
        return 'BitString({})'.format(self.name)


class OctetString(Type):

    def __init__(self, name):
        super(OctetString, self).__init__(name, 'OCTET STRING')

    def encode(self, data, encoder):
        encoder.append_bytes(bytearray([len(data)]) + data)

    def decode(self, decoder):
        length = decoder.read_bytes(1)[0]

        return decoder.read_bytes(length)

    def __repr__(self):
        return 'OctetString({})'.format(self.name)


class PrintableString(Type):

    def __init__(self, name):
        super(PrintableString, self).__init__(name, 'PrintableString')

    def encode(self, data, encoder):
        raise NotImplementedError()

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'PrintableString({})'.format(self.name)


class UniversalString(Type):

    def __init__(self, name):
        super(UniversalString, self).__init__(name, 'UniversalString')

    def encode(self, data, encoder):
        raise NotImplementedError()

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'UniversalString({})'.format(self.name)


class VisibleString(Type):

    def __init__(self, name):
        super(VisibleString, self).__init__(name, 'VisibleString')

    def encode(self, data, encoder):
        raise NotImplementedError()

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'VisibleString({})'.format(self.name)


class GeneralString(Type):

    def __init__(self, name):
        super(GeneralString, self).__init__(name, 'GeneralString')

    def encode(self, data, encoder):
        raise NotImplementedError()

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'GeneralString({})'.format(self.name)


class UTF8String(Type):

    def __init__(self, name):
        super(UTF8String, self).__init__(name, 'UTF8String')

    def encode(self, data, encoder):
        raise NotImplementedError()

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'UTF8String({})'.format(self.name)


class BMPString(Type):

    def __init__(self, name):
        super(BMPString, self).__init__(name, 'BMPString')

    def encode(self, data, encoder):
        raise NotImplementedError()

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'BMPString({})'.format(self.name)


class UTCTime(Type):

    def __init__(self, name):
        super(UTCTime, self).__init__(name, 'UTCTime')

    def encode(self, data, encoder):
        raise NotImplementedError()

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'UTCTime({})'.format(self.name)


class GeneralizedTime(Type):

    def __init__(self, name):
        super(GeneralizedTime, self).__init__(name, 'GeneralizedTime')

    def encode(self, data, encoder):
        raise NotImplementedError()

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'GeneralizedTime({})'.format(self.name)


class TeletexString(Type):

    def __init__(self, name):
        super(TeletexString, self).__init__(name, 'TeletexString')

    def encode(self, data, encoder):
        raise NotImplementedError()

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'TeletexString({})'.format(self.name)


class ObjectIdentifier(Type):

    def __init__(self, name):
        super(ObjectIdentifier, self).__init__(name, 'OBJECT IDENTIFIER')

    def encode(self, data, encoder):
        identifiers = [int(identifier) for identifier in data.split('.')]

        first_subidentifier = (40 * identifiers[0] + identifiers[1])
        encoded_subidentifiers = self.encode_subidentifier(
            first_subidentifier)

        for identifier in identifiers[2:]:
            encoded_subidentifiers += self.encode_subidentifier(identifier)

        encoder.append_bytes([len(encoded_subidentifiers)])
        encoder.append_bytes(encoded_subidentifiers)

    def decode(self, decoder):
        length = decoder.read_bytes(1)[0]
        data = decoder.read_bytes(length)
        offset = 0
        subidentifier, offset = self.decode_subidentifier(data, offset)
        decoded = [subidentifier // 40, subidentifier % 40]

        while offset < length:
            subidentifier, offset = self.decode_subidentifier(data, offset)
            decoded.append(subidentifier)

        return '.'.join([str(v) for v in decoded])

    def encode_subidentifier(self, subidentifier):
        encoder = [subidentifier & 0x7f]
        subidentifier >>= 7

        while subidentifier > 0:
            encoder.append(0x80 | (subidentifier & 0x7f))
            subidentifier >>= 7

        return encoder[::-1]

    def decode_subidentifier(self, data, offset):
        decoded = 0

        while data[offset] & 0x80:
            decoded += (data[offset] & 0x7f)
            decoded <<= 7
            offset += 1

        decoded += data[offset]

        return decoded, offset + 1

    def __repr__(self):
        return 'ObjectIdentifier({})'.format(self.name)


class Choice(Type):

    def __init__(self, name, members):
        super(Choice, self).__init__(name, 'CHOICE')
        self.members = members
        self.number_of_bits = len('{:b}'.format(len(members) - 1))

    def encode(self, data, encoder):
        if not isinstance(data, tuple):
            raise EncodeError("expected tuple, but got '{}'".format(data))

        for i, member in enumerate(self.members):
            if member.name == data[0]:
                if len(self.members) > 1:
                    encoder.append_integer(i, self.number_of_bits)

                member.encode(data[1], encoder)
                return

        raise EncodeError(
            "Expected choices are {}, but got '{}'.".format(
                [member.name for member in self.members],
                data[0]))

    def decode(self, data, offset):
        for member in self.members:
            if isinstance(member, Choice):
                try:
                    decoded, offset = member.decode(data, offset)
                    return (member.name, decoded), offset
                except DecodeChoiceError:
                    pass
            elif member.tag == data[offset]:
                decoded, offset = member.decode(data, offset)
                return (member.name, decoded), offset

        raise DecodeChoiceError()

    def __repr__(self):
        return 'Choice({}, [{}])'.format(
            self.name,
            ', '.join([repr(member) for member in self.members]))


class Null(Type):

    def __init__(self, name):
        super(Null, self).__init__(name, 'NULL')

    def encode(self, _, _encoder):
        pass

    def decode(self, _):
        return None

    def __repr__(self):
        return 'Null({})'.format(self.name)


class Any(Type):

    def __init__(self, name):
        super(Any, self).__init__(name, 'ANY')

    def encode(self, _, encoder):
        raise NotImplementedError()

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'Any({})'.format(self.name)


class OpenType(Type):

    def __init__(self, name):
        super(OpenType, self).__init__(name, 'OpenType')

    def encode(self, data, encoder):
        encoder.append_bytes(bytearray([len(data)]) + data)

    def decode(self, decoder):
        raise NotImplementedError()

    def __repr__(self):
        return 'OpenType({})'.format(self.name)


class Enumerated(Type):

    def __init__(self, name, values):
        super(Enumerated, self).__init__(name, 'ENUMERATED')
        values = enum_values_as_dict(values)
        self.values = values
        self.lowest_value = min(values)
        highest_value = max(values.keys()) - self.lowest_value
        self.number_of_bits = len('{:b}'.format(highest_value))

    def encode(self, data, encoder):
        for value, name in self.values.items():
            if data == name:
                encoder.append_integer(value - self.lowest_value,
                                       self.number_of_bits)
                return

        raise EncodeError(
            "Enumeration value '{}' not found in {}.".format(
                data,
                [value for value in self.values.values()]))

    def decode(self, decoder):
        value = decoder.read_integer(self.number_of_bits)

        return self.values[value + self.lowest_value]

    def __repr__(self):
        return 'Enumerated({})'.format(self.name)


class Recursive(Type):

    def __init__(self, name, type_name, module_name):
        super(Recursive, self).__init__(name, 'RECURSIVE')
        self.type_name = type_name
        self.module_name = module_name

    def encode(self, _data, _encoder):
        raise NotImplementedError(
            "Recursive types are not yet implemented (type '{}').".format(
                self.type_name))

    def decode(self, _decoder):
        raise NotImplementedError(
            "Recursive types are not yet implemented (type '{}').".format(
                self.type_name))

    def __repr__(self):
        return 'Recursive({})'.format(self.name)


class CompiledType(object):

    def __init__(self, type_):
        self._type = type_

    def encode(self, data):
        encoder = Encoder()
        self._type.encode(data, encoder)
        return encoder.as_bytearray()

    def decode(self, data):
        decoder = Decoder(bytearray(data))
        return self._type.decode(decoder)

    def __repr__(self):
        return repr(self._type)


class Compiler(compiler.Compiler):

    def process_type(self, type_name, type_descriptor, module_name):
        return CompiledType(self.compile_type(type_name,
                                              type_descriptor,
                                              module_name))

    def compile_type(self, name, type_descriptor, module_name):
        if '.' in type_descriptor['type']:
            type_descriptor = self.convert_class_member_type(type_descriptor,
                                                             module_name)

        type_name = type_descriptor['type']

        if type_name == 'SEQUENCE':
            compiled = Sequence(
                name,
                self.compile_members(type_descriptor['members'],
                                     module_name))
        elif type_name == 'SEQUENCE OF':
            compiled = SequenceOf(name,
                                  self.compile_type('',
                                                    type_descriptor['element'],
                                                    module_name))
        elif type_name == 'SET':
            compiled = Set(
                name,
                self.compile_members(type_descriptor['members'],
                                     module_name))
        elif type_name == 'SET OF':
            compiled = SetOf(name,
                             self.compile_type('',
                                               type_descriptor['element'],
                                               module_name))
        elif type_name == 'CHOICE':
            compiled = Choice(
                name,
                self.compile_members(type_descriptor['members'],
                                     module_name))
        elif type_name == 'INTEGER':
            compiled = Integer(name,
                               *self.get_restricted_to_range(type_descriptor,
                                                             module_name))
        elif type_name == 'REAL':
            compiled = Real(name)
        elif type_name == 'ENUMERATED':
            compiled = Enumerated(name, type_descriptor['values'])
        elif type_name == 'BOOLEAN':
            compiled = Boolean(name)
        elif type_name == 'OBJECT IDENTIFIER':
            compiled = ObjectIdentifier(name)
        elif type_name == 'OCTET STRING':
            compiled = OctetString(name)
        elif type_name == 'TeletexString':
            compiled = TeletexString(name)
        elif type_name == 'NumericString':
            compiled = NumericString(name)
        elif type_name == 'PrintableString':
            compiled = PrintableString(name)
        elif type_name == 'IA5String':
            compiled = IA5String(name)
        elif type_name == 'VisibleString':
            compiled = VisibleString(name)
        elif type_name == 'GeneralString':
            compiled = GeneralString(name)
        elif type_name == 'UTF8String':
            compiled = UTF8String(name)
        elif type_name == 'BMPString':
            compiled = BMPString(name)
        elif type_name == 'UTCTime':
            compiled = UTCTime(name)
        elif type_name == 'UniversalString':
            compiled = UniversalString(name)
        elif type_name == 'GeneralizedTime':
            compiled = GeneralizedTime(name)
        elif type_name == 'BIT STRING':
            size = type_descriptor.get('size', None)
            compiled = BitString(name, size)
        elif type_name == 'ANY':
            compiled = Any(name)
        elif type_name == 'ANY DEFINED BY':
            compiled = Any(name)
        elif type_name == 'NULL':
            compiled = Null(name)
        elif type_name == 'OpenType':
            compiled = OpenType(name)
        else:
            if type_name in self.types_backtrace:
                compiled = Recursive(name,
                                     type_name,
                                     module_name)
            else:
                self.types_backtrace_push(type_name)
                compiled = self.compile_type(
                    name,
                    *self.lookup_type_descriptor(
                        type_name,
                        module_name))
                self.types_backtrace_pop()

        return compiled

    def compile_members(self, members, module_name):
        compiled_members = []

        for member in members:
            if member == '...':
                continue

            if isinstance(member, list):
                compiled_members.extend(self.compile_members(member,
                                                             module_name))
                continue

            compiled_member = self.compile_type(member['name'],
                                                member,
                                                module_name)

            if 'optional' in member:
                compiled_member.optional = member['optional']

            if 'default' in member:
                compiled_member.default = member['default']

            compiled_members.append(compiled_member)

        return compiled_members


def compile_dict(specification):
    return Compiler(specification).process()


def decode_length(_data):
    raise DecodeError('Decode length is not supported for this codec.')



"""Distinguished Encoding Rules (DER) codec.

"""

from . import EncodeError
from . import DecodeError
from . import DecodeTagError
from . import DecodeContentsLengthError
from . import compiler
from .ber import Class
from .ber import Encoding
from .ber import Tag
from .ber import encode_length_definite
from .ber import decode_length_definite
from .ber import encode_signed_integer
from .ber import decode_signed_integer
from .ber import encode_tag
from .ber import decode_tag
from .ber import Boolean
from .ber import Enumerated
from .ber import Null
from .ber import ObjectIdentifier


class DecodeChoiceError(Exception):
    pass


class Type(object):

    def __init__(self, name, type_name, number, flags=0):
        self.name = name
        self.type_name = type_name

        if number is None:
            self.tag = None
        else:
            self.tag = encode_tag(number, flags)

        self.optional = False
        self.default = None

    def set_tag(self, number, flags):
        if not Class.APPLICATION & flags:
            flags |= Class.CONTEXT_SPECIFIC

        self.tag = encode_tag(number, flags)

    def decode_tag(self, data, offset):
        end_offset = offset + len(self.tag)

        if data[offset:end_offset] != self.tag:
            raise DecodeTagError(self.type_name,
                                 self.tag,
                                 data[offset:end_offset],
                                 offset)

        return end_offset


class Integer(Type):

    def __init__(self, name):
        super(Integer, self).__init__(name,
                                      'INTEGER',
                                      Tag.INTEGER)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_signed_integer(data))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return decode_signed_integer(data[offset:end_offset]), end_offset

    def __repr__(self):
        return 'Integer({})'.format(self.name)


class Real(Type):

    def __init__(self, name):
        super(Real, self).__init__(name, 'REAL', Tag.REAL)

    def encode(self, data, encoded):
        raise NotImplementedError()

    def decode(self, data, offset):
        raise NotImplementedError()

    def __repr__(self):
        return 'Real({})'.format(self.name)


class IA5String(Type):

    def __init__(self, name):
        super(IA5String, self).__init__(name,
                                        'IA5String',
                                        Tag.IA5_STRING)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('ascii'))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[offset:end_offset].decode('ascii'), end_offset

    def __repr__(self):
        return 'IA5String({})'.format(self.name)


class NumericString(Type):

    def __init__(self, name):
        super(NumericString, self).__init__(name,
                                            'NumericString',
                                            Tag.NUMERIC_STRING)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('ascii'))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[offset:end_offset].decode('ascii'), end_offset

    def __repr__(self):
        return 'NumericString({})'.format(self.name)


class MembersType(Type):

    def __init__(self, name, tag_name, tag, members):
        super(MembersType, self).__init__(name,
                                          tag_name,
                                          tag,
                                          Encoding.CONSTRUCTED)
        self.members = members

    def set_tag(self, number, flags):
        super(MembersType, self).set_tag(number,
                                         flags | Encoding.CONSTRUCTED)

    def encode(self, data, encoded):
        encoded_members = bytearray()

        for member in self.members:
            name = member.name

            if name in data:
                value = data[name]

                if isinstance(member, AnyDefinedBy):
                    member.encode(value, encoded_members, data)
                elif member.default != value or isinstance(member, Null):
                    member.encode(value, encoded_members)
            elif member.optional:
                pass
            elif member.default is None:
                raise EncodeError(
                    "member '{}' not found in {}".format(
                        name,
                        data))

        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(encoded_members)))
        encoded.extend(encoded_members)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        values = {}

        for member in self.members:
            try:
                if offset < end_offset:
                    if isinstance(member, AnyDefinedBy):
                        value, offset = member.decode(data, offset, values)
                    else:
                        value, offset = member.decode(data, offset)
                else:
                    raise IndexError
            except (DecodeError, IndexError) as e:
                if member.optional:
                    continue

                if member.default is None:
                    if isinstance(e, IndexError):
                        e = DecodeError('out of data at offset {}'.format(offset))

                    e.location.append(member.name)
                    raise e

                value = member.default

            values[member.name] = value

        return values, offset

    def __repr__(self):
        return '{}({}, [{}])'.format(
            self.__class__.__name__,
            self.name,
            ', '.join([repr(member) for member in self.members]))


class Sequence(MembersType):

    def __init__(self, name, members):
        super(Sequence, self).__init__(name, 'SEQUENCE', Tag.SEQUENCE, members)


class Set(MembersType):

    def __init__(self, name, members):
        super(Set, self).__init__(name, 'SET', Tag.SET, members)


class ArrayType(Type):

    def __init__(self, name, tag_name, tag, element_type):
        super(ArrayType, self).__init__(name,
                                        tag_name,
                                        tag,
                                        Encoding.CONSTRUCTED)
        self.element_type = element_type

    def set_tag(self, number, flags):
        super(ArrayType, self).set_tag(number,
                                       flags | Encoding.CONSTRUCTED)

    def encode(self, data, encoded):
        encoded_elements = bytearray()

        for entry in data:
            self.element_type.encode(entry, encoded_elements)

        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(encoded_elements)))
        encoded.extend(encoded_elements)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        decoded = []
        start_offset = offset

        while (offset - start_offset) < length:
            decoded_element, offset = self.element_type.decode(data, offset)
            decoded.append(decoded_element)

        return decoded, offset

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__,
                                   self.name,
                                   self.element_type)


class SequenceOf(ArrayType):

    def __init__(self, name, element_type):
        super(SequenceOf, self).__init__(name,
                                         'SEQUENCE OF',
                                         Tag.SEQUENCE,
                                         element_type)


class SetOf(ArrayType):

    def __init__(self, name, element_type):
        super(SetOf, self).__init__(name,
                                    'SET OF',
                                    Tag.SET,
                                    element_type)


class BitString(Type):

    def __init__(self, name):
        super(BitString, self).__init__(name,
                                        'BIT STRING',
                                        Tag.BIT_STRING)

    def encode(self, data, encoded):
        number_of_unused_bits = (8 - (data[1] % 8))

        if number_of_unused_bits == 8:
            number_of_unused_bits = 0

        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data[0]) + 1))
        encoded.append(number_of_unused_bits)
        encoded.extend(data[0])

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        number_of_bits = 8 * (length - 1) - data[offset]
        offset += 1

        return (bytearray(data[offset:end_offset]), number_of_bits), end_offset

    def __repr__(self):
        return 'BitString({})'.format(self.name)


class OctetString(Type):

    def __init__(self, name):
        super(OctetString, self).__init__(name,
                                          'OCTET STRING',
                                          Tag.OCTET_STRING)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return bytearray(data[offset:end_offset]), end_offset

    def __repr__(self):
        return 'OctetString({})'.format(self.name)


class PrintableString(Type):

    def __init__(self, name):
        super(PrintableString, self).__init__(name,
                                              'PrintableString',
                                              Tag.PRINTABLE_STRING)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('ascii'))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[offset:end_offset].decode('ascii'), end_offset

    def __repr__(self):
        return 'PrintableString({})'.format(self.name)


class UniversalString(Type):

    def __init__(self, name):
        super(UniversalString, self).__init__(name,
                                              'UniversalString',
                                              Tag.UNIVERSAL_STRING)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('ascii'))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[offset:end_offset].decode('ascii'), end_offset

    def __repr__(self):
        return 'UniversalString({})'.format(self.name)


class VisibleString(Type):

    def __init__(self, name):
        super(VisibleString, self).__init__(name,
                                            'VisibleString',
                                            Tag.VISIBLE_STRING)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('ascii'))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[offset:end_offset].decode('ascii'), end_offset

    def __repr__(self):
        return 'VisibleString({})'.format(self.name)


class GeneralString(Type):

    def __init__(self, name):
        super(GeneralString, self).__init__(name,
                                            'GeneralString',
                                            Tag.GENERAL_STRING)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('ascii'))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[offset:end_offset].decode('ascii'), end_offset

    def __repr__(self):
        return 'GeneralString({})'.format(self.name)


class UTF8String(Type):

    def __init__(self, name):
        super(UTF8String, self).__init__(name,
                                         'UTF8String',
                                         Tag.UTF8_STRING)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('utf-8'))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[offset:end_offset].decode('utf-8'), end_offset

    def __repr__(self):
        return 'UTF8String({})'.format(self.name)


class BMPString(Type):

    def __init__(self, name):
        super(BMPString, self).__init__(name,
                                        'BMPString',
                                        Tag.BMP_STRING)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return bytearray(data[offset:end_offset]), end_offset

    def __repr__(self):
        return 'BMPString({})'.format(self.name)


class GraphicString(Type):

    def __init__(self, name):
        super(GraphicString, self).__init__(name,
                                            'GraphicString',
                                            Tag.GRAPHIC_STRING)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('latin-1'))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[offset:end_offset].decode('latin-1'), end_offset

    def __repr__(self):
        return 'GraphicString({})'.format(self.name)


class UTCTime(Type):

    def __init__(self, name):
        super(UTCTime, self).__init__(name,
                                      'UTCTime',
                                      Tag.UTC_TIME)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.append(13)
        encoded.extend(bytearray((data + 'Z').encode('ascii')))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[offset:end_offset][:-1].decode('ascii'), end_offset

    def __repr__(self):
        return 'UTCTime({})'.format(self.name)


class GeneralizedTime(Type):

    def __init__(self, name):
        super(GeneralizedTime, self).__init__(name,
                                              'GeneralizedTime',
                                              Tag.GENERALIZED_TIME)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.append(len(data))
        encoded.extend(data.encode('ascii'))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[offset:end_offset].decode('ascii'), end_offset

    def __repr__(self):
        return 'GeneralizedTime({})'.format(self.name)


class TeletexString(Type):

    def __init__(self, name):
        super(TeletexString, self).__init__(name,
                                            'TeletexString',
                                            Tag.T61_STRING)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return bytearray(data[offset:end_offset]), end_offset

    def __repr__(self):
        return 'TeletexString({})'.format(self.name)


class Choice(Type):

    def __init__(self, name, members):
        super(Choice, self).__init__(name, 'CHOICE', None)
        self.members = members

    def set_tag(self, number, flags):
        super(Choice, self).set_tag(number,
                                    flags | Encoding.CONSTRUCTED)

    def encode(self, data, encoded):
        for member in self.members:
            if member.name == data[0]:
                member.encode(data[1], encoded)

                return

        raise EncodeError(
            "expected choices are {}, but got '{}'".format(
                [member.name for member in self.members],
                data[0]))

    def decode(self, data, offset):
        for member in self.members:
            if (isinstance(member, Choice)
                or member.tag == data[offset:offset + len(member.tag)]):
                try:
                    decoded, offset = member.decode(data, offset)
                except DecodeChoiceError:
                    pass
                else:
                    return (member.name, decoded), offset

        raise DecodeChoiceError()

    def __repr__(self):
        return 'Choice({}, [{}])'.format(
            self.name,
            ', '.join([repr(member) for member in self.members]))


class Any(Type):

    def __init__(self, name):
        super(Any, self).__init__(name, 'ANY', None)

    def encode(self, data, encoded):
        encoded.extend(data)

    def decode(self, data, offset):
        start = offset
        _, _, offset = decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[start:end_offset], end_offset

    def __repr__(self):
        return 'Any({})'.format(self.name)


class AnyDefinedBy(Type):

    def __init__(self, name, type_member, choices):
        super(AnyDefinedBy, self).__init__(name,
                                           'ANY DEFINED BY',
                                           None,
                                           None)
        self.type_member = type_member
        self.choices = choices

    def encode(self, data, encoded, values):
        if self.choices:
            self.choices[values[self.type_member]].encode(data, encoded)
        else:
            encoded.extend(data)

    def decode(self, data, offset, values):
        if self.choices:
            return self.choices[values[self.type_member]].decode(data,
                                                                 offset)
        else:
            start = offset
            _, _, offset = decode_tag(data, offset)
            length, offset = decode_length_definite(data, offset)
            end_offset = offset + length

            return data[start:end_offset], end_offset

    def __repr__(self):
        return 'AnyDefinedBy({})'.format(self.name)


class ExplicitTag(Type):

    def __init__(self, name, inner):
        super(ExplicitTag, self).__init__(name, 'Tag', None)
        self.inner = inner

    def set_tag(self, number, flags):
        super(ExplicitTag, self).set_tag(number,
                                         flags | Encoding.CONSTRUCTED)

    def encode(self, data, encoded):
        encoded_inner = bytearray()
        self.inner.encode(data, encoded_inner)
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(encoded_inner)))
        encoded.extend(encoded_inner)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        _, offset = decode_length_definite(data, offset)
        return self.inner.decode(data, offset)

    def __repr__(self):
        return 'Tag()'


class Recursive(Type):

    def __init__(self, name, type_name, module_name):
        super(Recursive, self).__init__(name, 'RECURSIVE', None)
        self.type_name = type_name
        self.module_name = module_name

    def encode(self, _data, _encoded):
        raise NotImplementedError(
            "Recursive types are not yet implemented (type '{}').".format(
                self.type_name))

    def decode(self, _data, _offset):
        raise NotImplementedError(
            "Recursive types are not yet implemented (type '{}').".format(
                self.type_name))

    def __repr__(self):
        return 'Recursive({})'.format(self.name)


class CompiledType(object):

    def __init__(self, type_):
        self._type = type_

    def encode(self, data):
        encoded = bytearray()
        self._type.encode(data, encoded)
        return encoded

    def decode(self, data):
        return self._type.decode(bytearray(data), 0)[0]

    def __repr__(self):
        return repr(self._type)


class Compiler(compiler.Compiler):

    def process_type(self, type_name, type_descriptor, module_name):
        return CompiledType(self.compile_type(type_name,
                                              type_descriptor,
                                              module_name))

    def compile_implicit_type(self, name, type_descriptor, module_name):
        type_name = type_descriptor['type']

        if type_name == 'SEQUENCE':
            compiled = Sequence(
                name,
                self.compile_members(type_descriptor['members'],
                                     module_name))
        elif type_name == 'SEQUENCE OF':
            compiled = SequenceOf(name,
                                  self.compile_type('',
                                                    type_descriptor['element'],
                                                    module_name))
        elif type_name == 'SET':
            compiled = Set(
                name,
                self.compile_members(type_descriptor['members'],
                                     module_name))
        elif type_name == 'SET OF':
            compiled = SetOf(name,
                             self.compile_type('',
                                               type_descriptor['element'],
                                               module_name))
        elif type_name == 'CHOICE':
            compiled = Choice(
                name,
                self.compile_members(type_descriptor['members'],
                                     module_name))
        elif type_name == 'INTEGER':
            compiled = Integer(name)
        elif type_name == 'REAL':
            compiled = Real(name)
        elif type_name == 'ENUMERATED':
            compiled = Enumerated(name, type_descriptor['values'])
        elif type_name == 'BOOLEAN':
            compiled = Boolean(name)
        elif type_name == 'OBJECT IDENTIFIER':
            compiled = ObjectIdentifier(name)
        elif type_name == 'OCTET STRING':
            compiled = OctetString(name)
        elif type_name == 'TeletexString':
            compiled = TeletexString(name)
        elif type_name == 'NumericString':
            compiled = NumericString(name)
        elif type_name == 'PrintableString':
            compiled = PrintableString(name)
        elif type_name == 'IA5String':
            compiled = IA5String(name)
        elif type_name == 'VisibleString':
            compiled = VisibleString(name)
        elif type_name == 'GeneralString':
            compiled = GeneralString(name)
        elif type_name == 'UTF8String':
            compiled = UTF8String(name)
        elif type_name == 'BMPString':
            compiled = BMPString(name)
        elif type_name == 'GraphicString':
            compiled = GraphicString(name)
        elif type_name == 'UTCTime':
            compiled = UTCTime(name)
        elif type_name == 'UniversalString':
            compiled = UniversalString(name)
        elif type_name == 'GeneralizedTime':
            compiled = GeneralizedTime(name)
        elif type_name == 'BIT STRING':
            compiled = BitString(name)
        elif type_name == 'ANY':
            compiled = Any(name)
        elif type_name == 'ANY DEFINED BY':
            choices = {}

            for key, value in type_descriptor['choices'].items():
                choices[key] = self.compile_type(key,
                                                 value,
                                                 module_name)

            compiled = AnyDefinedBy(name,
                                    type_descriptor['value'],
                                    choices)
        elif type_name == 'NULL':
            compiled = Null(name)
        else:
            if type_name in self.types_backtrace:
                compiled = Recursive(name,
                                     type_name,
                                     module_name)
            else:
                self.types_backtrace_push(type_name)
                compiled = self.compile_type(
                    name,
                    *self.lookup_type_descriptor(
                        type_name,
                        module_name))
                self.types_backtrace_pop()

        return compiled

    def compile_type(self, name, type_descriptor, module_name):
        compiled = self.compile_implicit_type(name,
                                              type_descriptor,
                                              module_name)

        if self.is_explicit_tag(type_descriptor):
            compiled = ExplicitTag(name, compiled)

        # Set any given tag.
        if 'tag' in type_descriptor:
            tag = type_descriptor['tag']
            class_ = tag.get('class', None)

            if class_ == 'APPLICATION':
                flags = Class.APPLICATION
            elif class_ == 'PRIVATE':
                flags = Class.PRIVATE
            else:
                flags = 0

            compiled.set_tag(tag['number'], flags)

        return compiled

    def compile_members(self, members, module_name):
        compiled_members = []

        for member in members:
            if member == '...':
                continue

            if isinstance(member, list):
                compiled_members.extend(self.compile_members(member,
                                                             module_name))
                continue

            compiled_member = self.compile_type(member['name'],
                                                member,
                                                module_name)

            if 'optional' in member:
                compiled_member.optional = member['optional']

            if 'default' in member:
                compiled_member.default = member['default']

            compiled_members.append(compiled_member)

        return compiled_members


def compile_dict(specification):
    return Compiler(specification).process()


def decode_length(data):
    try:
        return sum(decode_length_definite(bytearray(data), 1))
    except DecodeContentsLengthError as e:
        return (e.length + e.offset)
    except IndexError:
        return None



EXPECTED = {'Lightweight-Directory-Access-Protocol-V3': {'extensibility-implied': True,
                                              'imports': {},
                                              'object-classes': {},
                                              'object-sets': {},
                                              'tags': 'IMPLICIT',
                                              'types': {'AbandonRequest': {'tag': {'class': 'APPLICATION',
                                                                                   'number': 16},
                                                                           'type': 'MessageID'},
                                                        'AddRequest': {'members': [{'name': 'entry',
                                                                                    'type': 'LDAPDN'},
                                                                                   {'name': 'attributes',
                                                                                    'type': 'AttributeList'}],
                                                                       'tag': {'class': 'APPLICATION',
                                                                               'number': 8},
                                                                       'type': 'SEQUENCE'},
                                                        'AddResponse': {'tag': {'class': 'APPLICATION',
                                                                                'number': 9},
                                                                        'type': 'LDAPResult'},
                                                        'AssertionValue': {'type': 'OCTET '
                                                                                   'STRING'},
                                                        'Attribute': {'type': 'PartialAttribute'},
                                                        'AttributeDescription': {'type': 'LDAPString'},
                                                        'AttributeList': {'element': {'type': 'Attribute'},
                                                                          'type': 'SEQUENCE '
                                                                                  'OF'},
                                                        'AttributeSelection': {'element': {'type': 'LDAPString'},
                                                                               'type': 'SEQUENCE '
                                                                                       'OF'},
                                                        'AttributeValue': {'type': 'OCTET '
                                                                                   'STRING'},
                                                        'AttributeValueAssertion': {'members': [{'name': 'attributeDesc',
                                                                                                 'type': 'AttributeDescription'},
                                                                                                {'name': 'assertionValue',
                                                                                                 'type': 'AssertionValue'}],
                                                                                    'type': 'SEQUENCE'},
                                                        'AuthenticationChoice': {'members': [{'name': 'simple',
                                                                                              'tag': {'number': 0},
                                                                                              'type': 'OCTET '
                                                                                                      'STRING'},
                                                                                             {'name': 'sasl',
                                                                                              'tag': {'number': 3},
                                                                                              'type': 'SaslCredentials'},
                                                                                             '...'],
                                                                                 'type': 'CHOICE'},
                                                        'BindRequest': {'members': [{'name': 'version',
                                                                                     'restricted-to': [(1,
                                                                                                        127)],
                                                                                     'type': 'INTEGER'},
                                                                                    {'name': 'name',
                                                                                     'type': 'LDAPDN'},
                                                                                    {'name': 'authentication',
                                                                                     'type': 'AuthenticationChoice'}],
                                                                        'tag': {'class': 'APPLICATION',
                                                                                'number': 0},
                                                                        'type': 'SEQUENCE'},
                                                        'BindResponse': {'members': [{'components-of': 'LDAPResult'},
                                                                                     {'name': 'serverSaslCreds',
                                                                                      'optional': True,
                                                                                      'tag': {'number': 7},
                                                                                      'type': 'OCTET '
                                                                                              'STRING'}],
                                                                         'tag': {'class': 'APPLICATION',
                                                                                 'number': 1},
                                                                         'type': 'SEQUENCE'},
                                                        'CompareRequest': {'members': [{'name': 'entry',
                                                                                        'type': 'LDAPDN'},
                                                                                       {'name': 'ava',
                                                                                        'type': 'AttributeValueAssertion'}],
                                                                           'tag': {'class': 'APPLICATION',
                                                                                   'number': 14},
                                                                           'type': 'SEQUENCE'},
                                                        'CompareResponse': {'tag': {'class': 'APPLICATION',
                                                                                    'number': 15},
                                                                            'type': 'LDAPResult'},
                                                        'Control': {'members': [{'name': 'controlType',
                                                                                 'type': 'LDAPOID'},
                                                                                {'default': False,
                                                                                 'name': 'criticality',
                                                                                 'type': 'BOOLEAN'},
                                                                                {'name': 'controlValue',
                                                                                 'optional': True,
                                                                                 'type': 'OCTET '
                                                                                         'STRING'}],
                                                                    'type': 'SEQUENCE'},
                                                        'Controls': {'element': {'type': 'Control'},
                                                                     'type': 'SEQUENCE '
                                                                             'OF'},
                                                        'DelRequest': {'tag': {'class': 'APPLICATION',
                                                                               'number': 10},
                                                                       'type': 'LDAPDN'},
                                                        'DelResponse': {'tag': {'class': 'APPLICATION',
                                                                                'number': 11},
                                                                        'type': 'LDAPResult'},
                                                        'ExtendedRequest': {'members': [{'name': 'requestName',
                                                                                         'tag': {'number': 0},
                                                                                         'type': 'LDAPOID'},
                                                                                        {'name': 'requestValue',
                                                                                         'optional': True,
                                                                                         'tag': {'number': 1},
                                                                                         'type': 'OCTET '
                                                                                                 'STRING'}],
                                                                            'tag': {'class': 'APPLICATION',
                                                                                    'number': 23},
                                                                            'type': 'SEQUENCE'},
                                                        'ExtendedResponse': {'members': [{'components-of': 'LDAPResult'},
                                                                                         {'name': 'responseName',
                                                                                          'optional': True,
                                                                                          'tag': {'number': 10},
                                                                                          'type': 'LDAPOID'},
                                                                                         {'name': 'responseValue',
                                                                                          'optional': True,
                                                                                          'tag': {'number': 11},
                                                                                          'type': 'OCTET '
                                                                                                  'STRING'}],
                                                                             'tag': {'class': 'APPLICATION',
                                                                                     'number': 24},
                                                                             'type': 'SEQUENCE'},
                                                        'Filter': {'members': [{'element': {'type': 'Filter'},
                                                                                'name': 'and',
                                                                                'size': [(1,
                                                                                          'MAX')],
                                                                                'tag': {'number': 0},
                                                                                'type': 'SET '
                                                                                        'OF'},
                                                                               {'element': {'type': 'Filter'},
                                                                                'name': 'or',
                                                                                'size': [(1,
                                                                                          'MAX')],
                                                                                'tag': {'number': 1},
                                                                                'type': 'SET '
                                                                                        'OF'},
                                                                               {'name': 'not',
                                                                                'tag': {'number': 2},
                                                                                'type': 'Filter'},
                                                                               {'name': 'equalityMatch',
                                                                                'tag': {'number': 3},
                                                                                'type': 'AttributeValueAssertion'},
                                                                               {'name': 'substrings',
                                                                                'tag': {'number': 4},
                                                                                'type': 'SubstringFilter'},
                                                                               {'name': 'greaterOrEqual',
                                                                                'tag': {'number': 5},
                                                                                'type': 'AttributeValueAssertion'},
                                                                               {'name': 'lessOrEqual',
                                                                                'tag': {'number': 6},
                                                                                'type': 'AttributeValueAssertion'},
                                                                               {'name': 'present',
                                                                                'tag': {'number': 7},
                                                                                'type': 'AttributeDescription'},
                                                                               {'name': 'approxMatch',
                                                                                'tag': {'number': 8},
                                                                                'type': 'AttributeValueAssertion'},
                                                                               {'name': 'extensibleMatch',
                                                                                'tag': {'number': 9},
                                                                                'type': 'MatchingRuleAssertion'},
                                                                               '...'],
                                                                   'type': 'CHOICE'},
                                                        'IntermediateResponse': {'members': [{'name': 'responseName',
                                                                                              'optional': True,
                                                                                              'tag': {'number': 0},
                                                                                              'type': 'LDAPOID'},
                                                                                             {'name': 'responseValue',
                                                                                              'optional': True,
                                                                                              'tag': {'number': 1},
                                                                                              'type': 'OCTET '
                                                                                                      'STRING'}],
                                                                                 'tag': {'class': 'APPLICATION',
                                                                                         'number': 25},
                                                                                 'type': 'SEQUENCE'},
                                                        'LDAPDN': {'type': 'LDAPString'},
                                                        'LDAPMessage': {'members': [{'name': 'messageID',
                                                                                     'type': 'MessageID'},
                                                                                    {'members': [{'name': 'bindRequest',
                                                                                                  'type': 'BindRequest'},
                                                                                                 {'name': 'bindResponse',
                                                                                                  'type': 'BindResponse'},
                                                                                                 {'name': 'unbindRequest',
                                                                                                  'type': 'UnbindRequest'},
                                                                                                 {'name': 'searchRequest',
                                                                                                  'type': 'SearchRequest'},
                                                                                                 {'name': 'searchResEntry',
                                                                                                  'type': 'SearchResultEntry'},
                                                                                                 {'name': 'searchResDone',
                                                                                                  'type': 'SearchResultDone'},
                                                                                                 {'name': 'searchResRef',
                                                                                                  'type': 'SearchResultReference'},
                                                                                                 {'name': 'modifyRequest',
                                                                                                  'type': 'ModifyRequest'},
                                                                                                 {'name': 'modifyResponse',
                                                                                                  'type': 'ModifyResponse'},
                                                                                                 {'name': 'addRequest',
                                                                                                  'type': 'AddRequest'},
                                                                                                 {'name': 'addResponse',
                                                                                                  'type': 'AddResponse'},
                                                                                                 {'name': 'delRequest',
                                                                                                  'type': 'DelRequest'},
                                                                                                 {'name': 'delResponse',
                                                                                                  'type': 'DelResponse'},
                                                                                                 {'name': 'modDNRequest',
                                                                                                  'type': 'ModifyDNRequest'},
                                                                                                 {'name': 'modDNResponse',
                                                                                                  'type': 'ModifyDNResponse'},
                                                                                                 {'name': 'compareRequest',
                                                                                                  'type': 'CompareRequest'},
                                                                                                 {'name': 'compareResponse',
                                                                                                  'type': 'CompareResponse'},
                                                                                                 {'name': 'abandonRequest',
                                                                                                  'type': 'AbandonRequest'},
                                                                                                 {'name': 'extendedReq',
                                                                                                  'type': 'ExtendedRequest'},
                                                                                                 {'name': 'extendedResp',
                                                                                                  'type': 'ExtendedResponse'},
                                                                                                 '...',
                                                                                                 {'name': 'intermediateResponse',
                                                                                                  'type': 'IntermediateResponse'}],
                                                                                     'name': 'protocolOp',
                                                                                     'type': 'CHOICE'},
                                                                                    {'name': 'controls',
                                                                                     'optional': True,
                                                                                     'tag': {'number': 0},
                                                                                     'type': 'Controls'}],
                                                                        'type': 'SEQUENCE'},
                                                        'LDAPOID': {'type': 'OCTET '
                                                                            'STRING'},
                                                        'LDAPResult': {'members': [{'name': 'resultCode',
                                                                                    'type': 'ENUMERATED',
                                                                                    'values': [('success',
                                                                                                0),
                                                                                               ('operationsError',
                                                                                                1),
                                                                                               ('protocolError',
                                                                                                2),
                                                                                               ('timeLimitExceeded',
                                                                                                3),
                                                                                               ('sizeLimitExceeded',
                                                                                                4),
                                                                                               ('compareFalse',
                                                                                                5),
                                                                                               ('compareTrue',
                                                                                                6),
                                                                                               ('authMethodNotSupported',
                                                                                                7),
                                                                                               ('strongerAuthRequired',
                                                                                                8),
                                                                                               ('referral',
                                                                                                10),
                                                                                               ('adminLimitExceeded',
                                                                                                11),
                                                                                               ('unavailableCriticalExtension',
                                                                                                12),
                                                                                               ('confidentialityRequired',
                                                                                                13),
                                                                                               ('saslBindInProgress',
                                                                                                14),
                                                                                               ('noSuchAttribute',
                                                                                                16),
                                                                                               ('undefinedAttributeType',
                                                                                                17),
                                                                                               ('inappropriateMatching',
                                                                                                18),
                                                                                               ('constraintViolation',
                                                                                                19),
                                                                                               ('attributeOrValueExists',
                                                                                                20),
                                                                                               ('invalidAttributeSyntax',
                                                                                                21),
                                                                                               ('noSuchObject',
                                                                                                32),
                                                                                               ('aliasProblem',
                                                                                                33),
                                                                                               ('invalidDNSyntax',
                                                                                                34),
                                                                                               ('aliasDereferencingProblem',
                                                                                                36),
                                                                                               ('inappropriateAuthentication',
                                                                                                48),
                                                                                               ('invalidCredentials',
                                                                                                49),
                                                                                               ('insufficientAccessRights',
                                                                                                50),
                                                                                               ('busy',
                                                                                                51),
                                                                                               ('unavailable',
                                                                                                52),
                                                                                               ('unwillingToPerform',
                                                                                                53),
                                                                                               ('loopDetect',
                                                                                                54),
                                                                                               ('namingViolation',
                                                                                                64),
                                                                                               ('objectClassViolation',
                                                                                                65),
                                                                                               ('notAllowedOnNonLeaf',
                                                                                                66),
                                                                                               ('notAllowedOnRDN',
                                                                                                67),
                                                                                               ('entryAlreadyExists',
                                                                                                68),
                                                                                               ('objectClassModsProhibited',
                                                                                                69),
                                                                                               ('affectsMultipleDSAs',
                                                                                                71),
                                                                                               ('other',
                                                                                                80),
                                                                                               '...']},
                                                                                   {'name': 'matchedDN',
                                                                                    'type': 'LDAPDN'},
                                                                                   {'name': 'diagnosticMessage',
                                                                                    'type': 'LDAPString'},
                                                                                   {'name': 'referral',
                                                                                    'optional': True,
                                                                                    'tag': {'number': 3},
                                                                                    'type': 'Referral'}],
                                                                       'type': 'SEQUENCE'},
                                                        'LDAPString': {'type': 'OCTET '
                                                                               'STRING'},
                                                        'MatchingRuleAssertion': {'members': [{'name': 'matchingRule',
                                                                                               'optional': True,
                                                                                               'tag': {'number': 1},
                                                                                               'type': 'MatchingRuleId'},
                                                                                              {'name': 'type',
                                                                                               'optional': True,
                                                                                               'tag': {'number': 2},
                                                                                               'type': 'AttributeDescription'},
                                                                                              {'name': 'matchValue',
                                                                                               'tag': {'number': 3},
                                                                                               'type': 'AssertionValue'},
                                                                                              {'default': False,
                                                                                               'name': 'dnAttributes',
                                                                                               'tag': {'number': 4},
                                                                                               'type': 'BOOLEAN'}],
                                                                                  'type': 'SEQUENCE'},
                                                        'MatchingRuleId': {'type': 'LDAPString'},
                                                        'MessageID': {'restricted-to': [(0,
                                                                                         'maxInt')],
                                                                      'type': 'INTEGER'},
                                                        'ModifyDNRequest': {'members': [{'name': 'entry',
                                                                                         'type': 'LDAPDN'},
                                                                                        {'name': 'newrdn',
                                                                                         'type': 'RelativeLDAPDN'},
                                                                                        {'name': 'deleteoldrdn',
                                                                                         'type': 'BOOLEAN'},
                                                                                        {'name': 'newSuperior',
                                                                                         'optional': True,
                                                                                         'tag': {'number': 0},
                                                                                         'type': 'LDAPDN'}],
                                                                            'tag': {'class': 'APPLICATION',
                                                                                    'number': 12},
                                                                            'type': 'SEQUENCE'},
                                                        'ModifyDNResponse': {'tag': {'class': 'APPLICATION',
                                                                                     'number': 13},
                                                                             'type': 'LDAPResult'},
                                                        'ModifyRequest': {'members': [{'name': 'object',
                                                                                       'type': 'LDAPDN'},
                                                                                      {'element': {'members': [{'name': 'operation',
                                                                                                                'type': 'ENUMERATED',
                                                                                                                'values': [('add',
                                                                                                                            0),
                                                                                                                           ('delete',
                                                                                                                            1),
                                                                                                                           ('replace',
                                                                                                                            2),
                                                                                                                           '...']},
                                                                                                               {'name': 'modification',
                                                                                                                'type': 'PartialAttribute'}],
                                                                                                   'type': 'SEQUENCE'},
                                                                                       'name': 'changes',
                                                                                       'type': 'SEQUENCE '
                                                                                               'OF'}],
                                                                          'tag': {'class': 'APPLICATION',
                                                                                  'number': 6},
                                                                          'type': 'SEQUENCE'},
                                                        'ModifyResponse': {'tag': {'class': 'APPLICATION',
                                                                                   'number': 7},
                                                                           'type': 'LDAPResult'},
                                                        'PartialAttribute': {'members': [{'name': 'type',
                                                                                          'type': 'AttributeDescription'},
                                                                                         {'element': {'type': 'AttributeValue'},
                                                                                          'name': 'vals',
                                                                                          'type': 'SET '
                                                                                                  'OF'}],
                                                                             'type': 'SEQUENCE'},
                                                        'PartialAttributeList': {'element': {'type': 'PartialAttribute'},
                                                                                 'type': 'SEQUENCE '
                                                                                         'OF'},
                                                        'Referral': {'element': {'type': 'URI'},
                                                                     'size': [(1,
                                                                               'MAX')],
                                                                     'type': 'SEQUENCE '
                                                                             'OF'},
                                                        'RelativeLDAPDN': {'type': 'LDAPString'},
                                                        'SaslCredentials': {'members': [{'name': 'mechanism',
                                                                                         'type': 'LDAPString'},
                                                                                        {'name': 'credentials',
                                                                                         'optional': True,
                                                                                         'type': 'OCTET '
                                                                                                 'STRING'}],
                                                                            'type': 'SEQUENCE'},
                                                        'SearchRequest': {'members': [{'name': 'baseObject',
                                                                                       'type': 'LDAPDN'},
                                                                                      {'name': 'scope',
                                                                                       'type': 'ENUMERATED',
                                                                                       'values': [('baseObject',
                                                                                                   0),
                                                                                                  ('singleLevel',
                                                                                                   1),
                                                                                                  ('wholeSubtree',
                                                                                                   2),
                                                                                                  '...']},
                                                                                      {'name': 'derefAliases',
                                                                                       'type': 'ENUMERATED',
                                                                                       'values': [('neverDerefAliases',
                                                                                                   0),
                                                                                                  ('derefInSearching',
                                                                                                   1),
                                                                                                  ('derefFindingBaseObj',
                                                                                                   2),
                                                                                                  ('derefAlways',
                                                                                                   3)]},
                                                                                      {'name': 'sizeLimit',
                                                                                       'restricted-to': [(0,
                                                                                                          'maxInt')],
                                                                                       'type': 'INTEGER'},
                                                                                      {'name': 'timeLimit',
                                                                                       'restricted-to': [(0,
                                                                                                          'maxInt')],
                                                                                       'type': 'INTEGER'},
                                                                                      {'name': 'typesOnly',
                                                                                       'type': 'BOOLEAN'},
                                                                                      {'name': 'filter',
                                                                                       'type': 'Filter'},
                                                                                      {'name': 'attributes',
                                                                                       'type': 'AttributeSelection'}],
                                                                          'tag': {'class': 'APPLICATION',
                                                                                  'number': 3},
                                                                          'type': 'SEQUENCE'},
                                                        'SearchResultDone': {'tag': {'class': 'APPLICATION',
                                                                                     'number': 5},
                                                                             'type': 'LDAPResult'},
                                                        'SearchResultEntry': {'members': [{'name': 'objectName',
                                                                                           'type': 'LDAPDN'},
                                                                                          {'name': 'attributes',
                                                                                           'type': 'PartialAttributeList'}],
                                                                              'tag': {'class': 'APPLICATION',
                                                                                      'number': 4},
                                                                              'type': 'SEQUENCE'},
                                                        'SearchResultReference': {'element': {'type': 'URI'},
                                                                                  'size': [(1,
                                                                                            'MAX')],
                                                                                  'tag': {'class': 'APPLICATION',
                                                                                          'number': 19},
                                                                                  'type': 'SEQUENCE '
                                                                                          'OF'},
                                                        'SubstringFilter': {'members': [{'name': 'type',
                                                                                         'type': 'AttributeDescription'},
                                                                                        {'element': {'members': [{'name': 'initial',
                                                                                                                  'tag': {'number': 0},
                                                                                                                  'type': 'AssertionValue'},
                                                                                                                 {'name': 'any',
                                                                                                                  'tag': {'number': 1},
                                                                                                                  'type': 'AssertionValue'},
                                                                                                                 {'name': 'final',
                                                                                                                  'tag': {'number': 2},
                                                                                                                  'type': 'AssertionValue'}],
                                                                                                     'type': 'CHOICE'},
                                                                                         'name': 'substrings',
                                                                                         'size': [(1,
                                                                                                   'MAX')],
                                                                                         'type': 'SEQUENCE '
                                                                                                 'OF'}],
                                                                            'type': 'SEQUENCE'},
                                                        'URI': {'type': 'LDAPString'},
                                                        'UnbindRequest': {'tag': {'class': 'APPLICATION',
                                                                                  'number': 2},
                                                                          'type': 'NULL'}},
                                              'values': {'maxInt': {'type': 'INTEGER',
                                                                    'value': 2147483647}}}}


EXPECTED = {'Bar': {'extensibility-implied': False,
         'imports': {},
         'object-classes': {},
         'object-sets': {},
         'tags': 'EXPLICIT',
         'types': {'AcceptTypes': {'members': [{'element': {'type': 'BIT '
                                                                    'STRING'},
                                                'name': 'standardTypes',
                                                'optional': True,
                                                'tag': {'kind': 'IMPLICIT',
                                                        'number': 0},
                                                'type': 'SEQUENCE OF'},
                                               {'element': {'type': 'OCTET '
                                                                    'STRING'},
                                                'name': 'otherTypes',
                                                'optional': True,
                                                'tag': {'kind': 'IMPLICIT',
                                                        'number': 1},
                                                'type': 'SEQUENCE OF'}],
                                   'tag': {'class': 'APPLICATION',
                                           'kind': 'IMPLICIT',
                                           'number': 1},
                                   'type': 'SEQUENCE'},
                   'GetRequest': {'members': [{'name': 'headerOnly',
                                               'type': 'BOOLEAN'},
                                              {'name': 'lock',
                                               'type': 'BOOLEAN'},
                                              {'name': 'acceptTypes',
                                               'optional': True,
                                               'type': 'AcceptTypes'},
                                              {'name': 'url',
                                               'type': 'OCTET STRING'}],
                                  'tag': {'class': 'APPLICATION',
                                          'kind': 'IMPLICIT',
                                          'number': 0},
                                  'type': 'SEQUENCE'}},
         'values': {}}}


EXPECTED = {'PKCS-10': {'extensibility-implied': False,
             'imports': {'AuthenticationFramework': ['ALGORITHM'],
                         'InformationFramework': ['ATTRIBUTE', 'Name'],
                         'UsefulDefinitions': ['informationFramework',
                                               'authenticationFramework']},
             'object-classes': {},
             'object-sets': {'CRIAttributes': {'class': 'ATTRIBUTE',
                                               'members': ['.']},
                             'PKInfoAlgorithms': {'class': 'ALGORITHM',
                                                  'members': ['.']},
                             'SignatureAlgorithms': {'class': 'ALGORITHM',
                                                     'members': ['.']}},
             'tags': 'IMPLICIT',
             'types': {'AlgorithmIdentifier': {'members': [{'name': 'algorithm',
                                                            'table': 'IOSet',
                                                            'type': 'ALGORITHM.&id'},
                                                           {'name': 'parameters',
                                                            'optional': True,
                                                            'table': ['IOSet',
                                                                      ['algorithm']],
                                                            'type': 'ALGORITHM.&Type'}],
                                               'type': 'SEQUENCE'},
                       'Attribute': {'members': [{'name': 'type',
                                                  'table': 'IOSet',
                                                  'type': 'ATTRIBUTE.&id'},
                                                 {'element': {'table': ['IOSet',
                                                                        ['type']],
                                                              'type': 'ATTRIBUTE.&Type'},
                                                  'name': 'values',
                                                  'size': [(1, 'MAX')],
                                                  'type': 'SET OF'}],
                                     'type': 'SEQUENCE'},
                       'Attributes': {'element': {'type': 'Attribute'},
                                      'type': 'SET OF'},
                       'CertificationRequest': {'members': [{'name': 'certificationRequestInfo',
                                                             'type': 'CertificationRequestInfo'},
                                                            {'name': 'signatureAlgorithm',
                                                             'type': 'AlgorithmIdentifier'},
                                                            {'name': 'signature',
                                                             'type': 'BIT '
                                                                     'STRING'}],
                                                'type': 'SEQUENCE'},
                       'CertificationRequestInfo': {'members': [{'name': 'version',
                                                                 'restricted-to': ['v1', '...'],
                                                                 'type': 'INTEGER'},
                                                                {'name': 'subject',
                                                                 'type': 'Name'},
                                                                {'name': 'subjectPKInfo',
                                                                 'type': 'SubjectPublicKeyInfo'},
                                                                {'name': 'attributes',
                                                                 'tag': {'number': 0},
                                                                 'type': 'Attributes'}],
                                                    'type': 'SEQUENCE'},
                       'SubjectPublicKeyInfo': {'members': [{'name': 'algorithm',
                                                             'type': 'AlgorithmIdentifier'},
                                                            {'name': 'subjectPublicKey',
                                                             'type': 'BIT '
                                                                     'STRING'}],
                                                'type': 'SEQUENCE'}},
             'values': {}}}



EXPECTED = {'ModuleTagsAutomatic': {'extensibility-implied': False,
                         'imports': {},
                         'object-classes': {},
                         'object-sets': {},
                         'tags': 'AUTOMATIC',
                         'types': {'A': {'tag': {'kind': 'IMPLICIT',
                                                 'number': 3},
                                         'type': 'INTEGER'},
                                   'AI': {'tag': {'kind': 'IMPLICIT',
                                                  'number': 3},
                                          'type': 'INTEGER'},
                                   'BA': {'tag': {'kind': 'IMPLICIT',
                                                  'number': 4},
                                          'type': 'A'},
                                   'BIA': {'tag': {'kind': 'IMPLICIT',
                                                   'number': 4},
                                           'type': 'A'},
                                   'BIAI': {'tag': {'kind': 'IMPLICIT',
                                                    'number': 4},
                                            'type': 'AI'},
                                   'C1': {'members': [{'members': [{'name': 'a',
                                                                    'tag': {'kind': 'IMPLICIT',
                                                                            'number': 0},
                                                                    'type': 'INTEGER'}],
                                                       'name': 'a',
                                                       'tag': {'kind': 'EXPLICIT',
                                                               'number': 0},
                                                       'type': 'CHOICE'}],
                                          'type': 'CHOICE'},
                                   'CBA': {'tag': {'kind': 'IMPLICIT',
                                                   'number': 5},
                                           'type': 'BA'},
                                   'CBIAI': {'tag': {'kind': 'IMPLICIT',
                                                     'number': 5},
                                             'type': 'BIAI'},
                                   'CIBIA': {'tag': {'kind': 'IMPLICIT',
                                                     'number': 5},
                                             'type': 'BIA'},
                                   'CIBIAI': {'tag': {'kind': 'IMPLICIT',
                                                      'number': 5},
                                              'type': 'BIAI'},
                                   'S1': {'members': [{'name': 'a',
                                                       'tag': {'kind': 'IMPLICIT',
                                                               'number': 0},
                                                       'type': 'INTEGER'},
                                                      {'name': 'b',
                                                       'optional': True,
                                                       'tag': {'kind': 'IMPLICIT',
                                                               'number': 1},
                                                       'type': 'BOOLEAN'}],
                                          'type': 'SEQUENCE'},
                                   'S2': {'members': [{'name': 'a',
                                                       'type': 'INTEGER'},
                                                      {'name': 'b',
                                                       'tag': {'kind': 'IMPLICIT',
                                                               'number': 2},
                                                       'type': 'S1'},
                                                      {'members': [{'name': 'a',
                                                                    'tag': {'kind': 'IMPLICIT',
                                                                            'number': 0},
                                                                    'type': 'BOOLEAN'}],
                                                       'name': 'c',
                                                       'type': 'CHOICE'}],
                                          'type': 'SEQUENCE'},
                                   'S3': {'members': [{'name': 'a',
                                                       'type': 'INTEGER'},
                                                      {'name': 'b',
                                                       'tag': {'kind': 'IMPLICIT',
                                                               'number': 2},
                                                       'type': 'S1'},
                                                      {'members': [{'name': 'a',
                                                                    'tag': {'kind': 'IMPLICIT',
                                                                            'number': 0},
                                                                    'type': 'BOOLEAN'}],
                                                       'name': 'c',
                                                       'tag': {'kind': 'EXPLICIT',
                                                               'number': 3},
                                                       'type': 'CHOICE'}],
                                          'type': 'SEQUENCE'},
                                   'S4': {'members': [{'name': 'a',
                                                       'type': 'INTEGER'},
                                                      {'name': 'b',
                                                       'tag': {'kind': 'EXPLICIT',
                                                               'number': 1},
                                                       'type': 'C1'},
                                                      {'name': 'c',
                                                       'tag': {'kind': 'IMPLICIT',
                                                               'number': 2},
                                                       'type': 'S1'},
                                                      {'members': [{'name': 'a',
                                                                    'tag': {'kind': 'IMPLICIT',
                                                                            'number': 0},
                                                                    'type': 'BOOLEAN'}],
                                                       'name': 'd',
                                                       'type': 'CHOICE'}],
                                          'type': 'SEQUENCE'},
                                   'S5': {'members': [{'name': 'a',
                                                       'tag': {'kind': 'IMPLICIT',
                                                               'number': 0},
                                                       'type': 'INTEGER'},
                                                      {'name': 'b',
                                                       'tag': {'kind': 'IMPLICIT',
                                                               'number': 1},
                                                       'type': 'S1'},
                                                      {'members': [{'name': 'a',
                                                                    'tag': {'kind': 'IMPLICIT',
                                                                            'number': 0},
                                                                    'type': 'BOOLEAN'}],
                                                       'name': 'c',
                                                       'tag': {'kind': 'EXPLICIT',
                                                               'number': 2},
                                                       'type': 'CHOICE'}],
                                          'type': 'SEQUENCE'},
                                   'S6': {'members': [{'name': 'a',
                                                       'tag': {'kind': 'IMPLICIT',
                                                               'number': 0},
                                                       'type': 'INTEGER'},
                                                      '...',
                                                      {'name': 'b',
                                                       'tag': {'kind': 'IMPLICIT',
                                                               'number': 1},
                                                       'type': 'BOOLEAN'}],
                                          'type': 'SEQUENCE'},
                                   'S7': {'members': [{'name': 'a',
                                                       'tag': {'kind': 'IMPLICIT',
                                                               'number': 2},
                                                       'type': 'INTEGER'},
                                                      '...',
                                                      {'name': 'b',
                                                       'type': 'BOOLEAN'}],
                                          'type': 'SEQUENCE'},
                                   'S8': {'members': [{'element': {'members': [{'name': 'a',
                                                                                'tag': {'kind': 'IMPLICIT',
                                                                                        'number': 0},
                                                                                'type': 'INTEGER'},
                                                                               {'name': 'b',
                                                                                'tag': {'kind': 'IMPLICIT',
                                                                                        'number': 1},
                                                                                'type': 'BOOLEAN'}],
                                                                   'type': 'CHOICE'},
                                                       'name': 'a',
                                                       'tag': {'kind': 'IMPLICIT',
                                                               'number': 0},
                                                       'type': 'SEQUENCE OF'}],
                                          'type': 'SEQUENCE'},
                                   'S9': {'members': [{'element': {'members': [{'name': 'a',
                                                                                'tag': {'kind': 'IMPLICIT',
                                                                                        'number': 0},
                                                                                'type': 'INTEGER'},
                                                                               {'name': 'b',
                                                                                'tag': {'kind': 'IMPLICIT',
                                                                                        'number': 1},
                                                                                'type': 'BOOLEAN'}],
                                                                   'type': 'CHOICE'},
                                                       'name': 'a',
                                                       'tag': {'kind': 'IMPLICIT',
                                                               'number': 0},
                                                       'type': 'SET OF'}],
                                          'type': 'SET'}},
                         'values': {}}}
